import datetime
import os
from collections import deque
from functools import partial, lru_cache
from numbers import Integral
from operator import attrgetter
from pathlib import Path
from typing import Any, Self
import pathlib

import aiocache
import pandas as pd
import sqlalchemy as sa
from aiocache import cached, Cache
from alembic import config, command
from sqlalchemy import Table, select
from sqlalchemy.orm import selectinload
from toolz import (
    concat,
    merge,
    partition_all,
)

from ziplime.assets.domain.asset_type import AssetType
from ziplime.assets.entities.asset import Asset
from ziplime.assets.entities.currency_symbol_mapping import CurrencySymbolMapping
from ziplime.assets.entities.equity_symbol_mapping import EquitySymbolMapping
from ziplime.assets.entities.symbol_universe import SymbolsUniverse
from ziplime.assets.models.asset_router import AssetRouter
from ziplime.assets.entities.commodity import Commodity
from ziplime.assets.entities.currency import Currency
from ziplime.assets.models.commodity_model import CommodityModel
from ziplime.assets.models.currency_model import CurrencyModel
from ziplime.assets.models.currency_symbol_mapping_model import CurrencySymbolMappingModel
from ziplime.assets.models.equity_model import EquityModel
from ziplime.assets.models.futures_contract_model import FuturesContractModel
from ziplime.assets.models.symbols_universe import SymbolsUniverseModel
from ziplime.assets.models.symbols_universe_asset import SymbolsUniverseAssetModel
from ziplime.trading.models.trading_pair import TradingPair
from ziplime.core.db.base_model import BaseModel
from ziplime.errors import (
    EquitiesNotFound,
    FutureContractsNotFound,
    MultipleSymbolsFound,
    SameSymbolUsedAcrossCountries,
    SidsNotFound,
    SymbolNotFound,
    NotAssetConvertible,
)
from ziplime.utils.functional import invert
from ziplime.utils.numpy_utils import as_column
from ziplime.utils.sqlite_utils import group_into_chunks, SQLITE_MAX_VARIABLE_NUMBER

from ziplime.assets.models.exchange_info import ExchangeInfo

import numpy as np
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine, async_sessionmaker

from ziplime.assets.models.asset_model import AssetModel
from ziplime.assets.domain.continuous_future import ContinuousFuture
from ziplime.assets.models.equity_symbol_mapping_model import EquitySymbolMappingModel
from ziplime.assets.entities.equity import Equity
from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.assets.domain.ordered_contracts import CHAIN_PREDICATES, OrderedContracts, ADJUSTMENT_STYLES
from ziplime.assets.repositories.asset_repository import AssetRepository
from ziplime.assets.utils import _convert_asset_timestamp_fields, _filter_future_kwargs, \
    _filter_equity_kwargs, _encode_continuous_future_sid, Lifetimes, \
    build_grouped_ownership_map, OwnershipPeriod, SYMBOL_COLUMNS, split_delimited_symbol


class SqlAlchemyAssetRepository(AssetRepository):
    """An AssetFinder is an interface to a database of Asset metadata written by
    an ``AssetDBWriter``.

    This class provides methods for looking up assets by unique integer id or
    by symbol.  For historical reasons, we refer to these unique ids as 'sids'.

    Parameters
    ----------
    engine : str or SQLAlchemy.engine
        An engine with a connection to the asset database to use, or a string
        that can be parsed by SQLAlchemy as a URI.
    future_chain_predicates : dict
        A dict mapping future root symbol to a predicate function which accepts
    a contract as a parameter and returns whether or not the contract should be
    included in the chain.

    See Also
    --------
    :class:`ziplime.assets.repositories.AssetsRepository`
    """

    def __init__(self, db_url: str, future_chain_predicates):
        self.db_url = db_url
        self._asset_cache = {}
        self._asset_type_cache = {}
        self._caches = (self._asset_cache, self._asset_type_cache)

        self._future_chain_predicates = (
            future_chain_predicates if future_chain_predicates is not None else {}
        )
        self._ordered_contracts = {}

        # Populated on first call to `lifetimes`.
        self._asset_lifetimes = {}
        self.migrate()

    # def get_database_path(self, bundle_name: str, bundle_version: str) -> Path:
    #     return Path(self._base_storage_path, "assets", f"{bundle_name}_{bundle_version}.sqlite")

    # def get_database_url_sync(self, bundle_name: str, bundle_version: str) -> str:
    #     return f"sqlite:////{self.get_database_path(bundle_name=bundle_name, bundle_version=bundle_version)}"
    #
    # def get_database_url_async(self, bundle_name: str, bundle_version: str) -> str:
    #     return f"sqlite+aiosqlite:////{self.get_database_path(bundle_name=bundle_name, bundle_version=bundle_version)}"

    @property
    @lru_cache
    def session_maker(self) -> async_sessionmaker[AsyncSession]:
        engine = create_async_engine(self.db_url, pool_pre_ping=True, pool_size=20)
        session_maker = async_sessionmaker(autocommit=False, autoflush=True, bind=engine, class_=AsyncSession,
                                           expire_on_commit=False)
        return session_maker

    async def add_all_and_commit(self, models: list[BaseModel]):
        async with self.session_maker() as session:
            session.add_all(models)
            await session.commit()

    async def save_trading_pairs(self, trading_pairs: list[TradingPair]) -> None:
        await self.add_all_and_commit(trading_pairs)

    async def save_asset_routers(self, asset_routers: list[AssetRouter]) -> None:
        await self.add_all_and_commit(asset_routers)

    async def save_currencies(self, currencies: list[Currency]) -> None:
        assets_db = []
        asset_routers = []
        symbol_mappings = []
        async with self.session_maker() as session:

            for currency in currencies:
                asset_router = AssetRouter(
                    sid=currency.sid,
                    asset_type=AssetType.CURRENCY.value
                )
                asset_routers.append(asset_router)
                session.add(asset_router)
                await session.commit()
                asset_db = CurrencyModel(
                    sid=asset_router.sid,
                    start_date=currency.start_date,
                    first_traded=currency.first_traded,
                    end_date=currency.end_date,
                    asset_name=currency.asset_name,
                    auto_close_date=currency.auto_close_date,
                    mic=currency.mic
                )
                session.add(asset_db)
                await session.commit()

                assets_db.append(asset_db)
                for symbol_mapping in currency.symbol_mapping.values():
                    exchange = await self.get_exchange_by_name(exchange_name=symbol_mapping.exchange_name)
                    if exchange is None:
                        raise ValueError(f"Exchange {symbol_mapping.exchange_name} not found. Please register it.")
                    symbol_mapping_model = CurrencySymbolMappingModel(
                        sid=asset_db.sid,
                        symbol=symbol_mapping.symbol,
                        start_date=symbol_mapping.start_date,
                        end_date=symbol_mapping.end_date,
                        exchange=exchange.exchange,
                    )
                    symbol_mappings.append(symbol_mapping_model)
                    session.add(symbol_mapping_model)
                    await session.commit()

            # trading_pair = TradingPair(
            #     id=uuid.uuid4(),
            #     base_asset_sid=asset_router.sid,
            #     quote_asset_sid=asset_db.sid,
            #     exchange=asset["exchange"],
            # )
            # trading_pairs.append(trading_pair)

        # do this in one transaction
        #     session.add_all(asset_routers)
        #     await session.commit()
        #     session.add_all(assets_db)
        #     session.add_all(symbol_mappings)
        #     await session.commit()

    async def save_symbol_universe(self, symbol_universe: SymbolsUniverse):
        async with self.session_maker() as session:
            symbol_universe_model = SymbolsUniverseModel(symbol=symbol_universe.symbol,
                                                         universe_type=symbol_universe.universe_type,
                                                         name=symbol_universe.name,
                                                         )
            session.add(symbol_universe_model)
            await session.commit()
            await session.refresh(symbol_universe_model)
            symbol_universe_assets = [
                SymbolsUniverseAssetModel(symbol_universe_id=symbol_universe_model.id, asset_sid=asset.sid)
                for asset in symbol_universe.assets
            ]
            session.add_all(symbol_universe_assets)
            await session.commit()

    @aiocache.cached(cache=Cache.MEMORY)
    async def get_exchange_by_name(self, exchange_name: str) -> ExchangeInfo | None:
        async with self.session_maker() as session:
            q = select(ExchangeInfo).where(ExchangeInfo.exchange == exchange_name)
            exchange = (await session.execute(q)).scalar_one_or_none()
            return exchange

    async def save_equities(self, equities: list[Equity]) -> None:
        assets_db = []
        asset_routers = []
        symbol_mappings = []
        for equity in equities:
            asset_router = AssetRouter(
                sid=equity.sid,
                asset_type=AssetType.EQUITY.value
            )
            asset_routers.append(asset_router)
        async with self.session_maker() as session:
            session.add_all(asset_routers)
            await session.commit()

        for i, equity in enumerate(equities):
            asset_db = EquityModel(
                sid=asset_routers[i].sid,
                start_date=equity.start_date,
                first_traded=equity.first_traded,
                end_date=equity.end_date,
                asset_name=equity.asset_name,
                auto_close_date=equity.auto_close_date,
                mic=equity.mic
            )
            assets_db.append(asset_db)
            for symbol_mapping in equity.symbol_mapping.values():
                exchange = await self.get_exchange_by_name(exchange_name=symbol_mapping.exchange_name)
                if exchange is None:
                    raise ValueError(f"Exchange {symbol_mapping.exchange_name} not found. Please register it.")
                equity_symbol_mapping = EquitySymbolMappingModel(
                    sid=asset_routers[i].sid,
                    company_symbol=symbol_mapping.company_symbol,
                    symbol=symbol_mapping.symbol,
                    share_class_symbol=symbol_mapping.share_class_symbol,
                    start_date=symbol_mapping.start_date,
                    end_date=symbol_mapping.end_date,
                    exchange=exchange.exchange,
                )
                symbol_mappings.append(equity_symbol_mapping)

                # trading_pair = TradingPair(
                #     id=uuid.uuid4(),
                #     base_asset_sid=asset_router.sid,
                #     quote_asset_sid=asset_db.sid,
                #     exchange=asset["exchange"],
                # )
                # trading_pairs.append(trading_pair)

        # do this in one transaction
        async with self.session_maker() as session:
            session.add_all(assets_db)
            session.add_all(symbol_mappings)
            await session.commit()

    async def save_exchanges(self, exchanges: list[ExchangeInfo]) -> None:
        await self.add_all_and_commit(exchanges)

    async def save_equity_symbol_mappings(self, equity_symbol_mappings: list[EquitySymbolMappingModel]) -> None:
        await self.add_all_and_commit(equity_symbol_mappings)

    @cached(cache=Cache.MEMORY)
    async def get_all_assets(self) -> dict[int, Asset]:
        async with self.session_maker() as session:
            q_equities = select(EquityModel).options(selectinload(EquityModel.asset_router)).options(
                selectinload(EquityModel.equity_symbol_mappings)
            )
            equities = list((await session.execute(q_equities)).scalars().all())

            q_futures_contracts = select(FuturesContractModel).options(selectinload(FuturesContractModel.asset_router))
            futures_contracts = list((await session.execute(q_futures_contracts)).scalars().all())

            q_currencies = select(CurrencyModel).options(selectinload(CurrencyModel.asset_router))
            currencies = list((await session.execute(q_currencies)).scalars().all())

            q_commodities = select(CommodityModel).options(selectinload(CommodityModel.asset_router))
            commodities = list((await session.execute(q_commodities)).scalars().all())

            res = {
                **{asset.sid: asset for asset in equities},
                **{asset.sid: asset for asset in futures_contracts},
                **{asset.sid: asset for asset in currencies},
                **{asset.sid: asset for asset in commodities},
            }
        return res

    @cached(cache=Cache.MEMORY)
    async def get_all_universes(self) -> dict[str, SymbolsUniverse]:
        async with self.session_maker() as session:
            q = select(SymbolsUniverseModel).options(
                selectinload(SymbolsUniverseModel.assets)
            )
            universes = list((await session.execute(q)).scalars())

            res = {
                universe.symbol:
                    SymbolsUniverse(
                        assets=await self.get_assets_by_sids(sids=[asset.asset_sid for asset in universe.assets]),
                        universe_type=universe.universe_type,
                        symbol=universe.symbol,
                        name=universe.name
                    )
                for
                universe in universes
            }
        return res

    @aiocache.cached(cache=Cache.MEMORY)
    async def get_asset_by_symbol(self, symbol: str, asset_type: AssetType,
                                  exchange_name: str | None) -> Asset | None:
        match asset_type:
            case AssetType.EQUITY:
                return await self.get_equity_by_symbol(symbol=symbol, exchange_name=exchange_name)
            case AssetType.FUTURES_CONTRACT:
                return await self.get_futures_contract_by_symbol(symbol=symbol, exchange_name=exchange_name)
            case AssetType.CURRENCY:
                return await self.get_currency_by_symbol(symbol=symbol)
            case _:
                raise ValueError(f"Invalid asset type: {asset_type}")

    async def get_asset_by_sid(self, sid: int) -> AssetModel | None:
        assets_by_sid = await self.get_all_assets()
        return assets_by_sid.get(sid, None)

    async def get_assets_by_sids(self, sids: list[int]) -> list[Asset]:
        assets_by_sid = await self.get_all_assets()
        return [assets_by_sid.get(sid, None) for sid in sids]

    async def get_symbols_universe(self, symbol: str) -> SymbolsUniverse | None:
        universes_by_symbol = await self.get_all_universes()
        return universes_by_symbol.get(symbol, None)

    @aiocache.cached(cache=Cache.MEMORY)
    async def get_currency_by_symbol(self, symbol: str, exchange_name: str) -> Currency | None:
        currencies = await self.get_currencies_by_symbols(symbols=[symbol], exchange_name=exchange_name)
        if currencies:
            return currencies[0]
        return None

    @aiocache.cached(cache=Cache.MEMORY)
    async def get_currencies_by_symbols(self, symbols: list[str], exchange_name: str) -> list[Currency]:
        async with self.session_maker() as session:
            q_currency_symbol_mapping = select(CurrencySymbolMappingModel).where(
                CurrencySymbolMappingModel.exchange == exchange_name,
                CurrencySymbolMappingModel.symbol.in_(symbols))

            currency_mappings = (await session.execute(q_currency_symbol_mapping)).scalars()

            q_currencies = select(CurrencyModel).where(
                CurrencyModel.sid.in_([currency_mapping.sid for currency_mapping in currency_mappings])).options(
                selectinload(CurrencyModel.asset_router)).options(selectinload(CurrencyModel.currency_symbol_mappings))
            assets: list[CurrencyModel] = list((await session.execute(q_currencies)).scalars())

            return [Currency(
                sid=asset.sid,
                asset_name=asset.asset_name,
                start_date=asset.start_date,
                first_traded=asset.first_traded,
                end_date=asset.end_date,
                auto_close_date=asset.auto_close_date,
                symbol_mapping={
                    currency_mapping.exchange: CurrencySymbolMapping(
                        symbol=currency_mapping.symbol,
                        exchange_name=currency_mapping.exchange,
                        end_date=currency_mapping.end_date,
                        start_date=currency_mapping.start_date
                    )
                    for currency_mapping in asset.currency_symbol_mappings
                },
                mic=asset.mic
            ) for asset in assets]

    async def get_commodity_by_symbol(self, symbol: str) -> Commodity | None:
        raise NotImplementedError("Not implemented")

    async def get_futures_contract_by_symbol(self, symbol: str, exchange_name: str) -> FuturesContract | None:
        raise NotImplementedError("Not implemented")

    async def get_equities_by_symbols_and_exchange(self, symbols: list[str], exchange_name: str) -> list[Equity]:
        async with self.session_maker() as session:
            q_equity_symbol_mapping = select(EquitySymbolMappingModel).where(
                EquitySymbolMappingModel.exchange == exchange_name,
                EquitySymbolMappingModel.symbol.in_(symbols))

            equity_mappings = (await session.execute(q_equity_symbol_mapping)).scalars()

            q_equities = select(EquityModel).where(
                EquityModel.sid.in_([equity_mapping.sid for equity_mapping in equity_mappings])).options(
                selectinload(EquityModel.asset_router)).options(selectinload(EquityModel.equity_symbol_mappings))
            assets: list[EquityModel] = list((await session.execute(q_equities)).scalars())

            return [Equity(
                sid=asset.sid,
                asset_name=asset.asset_name,
                start_date=asset.start_date,
                first_traded=asset.first_traded,
                end_date=asset.end_date,
                auto_close_date=asset.auto_close_date,
                symbol_mapping={
                    equity_mapping.exchange: EquitySymbolMapping(
                        company_symbol=equity_mapping.company_symbol,
                        symbol=equity_mapping.symbol,
                        exchange_name=equity_mapping.exchange,
                        share_class_symbol=equity_mapping.share_class_symbol,
                        end_date=equity_mapping.end_date,
                        start_date=equity_mapping.start_date
                    )
                    for equity_mapping in asset.equity_symbol_mappings
                },
                mic=asset.mic
            ) for asset in assets]

    async def get_equities_by_symbols(self, symbols: list[str]) -> list[Equity]:
        async with self.session_maker() as session:
            q_equity_symbol_mapping = select(EquitySymbolMappingModel).where(
                EquitySymbolMappingModel.symbol.in_(symbols))

            equity_mappings = (await session.execute(q_equity_symbol_mapping)).scalars()

            q_equities = select(EquityModel).where(
                EquityModel.sid.in_([equity_mapping.sid for equity_mapping in equity_mappings])).options(
                selectinload(EquityModel.asset_router)).options(selectinload(EquityModel.equity_symbol_mappings))
            assets: list[EquityModel] = list((await session.execute(q_equities)).scalars())

            return [Equity(
                sid=asset.sid,
                asset_name=asset.asset_name,
                start_date=asset.start_date,
                first_traded=asset.first_traded,
                end_date=asset.end_date,
                auto_close_date=asset.auto_close_date,
                symbol_mapping={
                    equity_mapping.exchange: EquitySymbolMapping(
                        company_symbol=equity_mapping.company_symbol,
                        symbol=equity_mapping.symbol,
                        exchange_name=equity_mapping.exchange,
                        share_class_symbol=equity_mapping.share_class_symbol,
                        end_date=equity_mapping.end_date,
                        start_date=equity_mapping.start_date
                    )
                    for equity_mapping in asset.equity_symbol_mappings
                },
                mic=asset.mic
            ) for asset in assets]

    async def get_equity_by_symbol(self, symbol: str, exchange_name: str) -> Equity | None:
        equities = await self.get_equities_by_symbols_and_exchange(symbols=[symbol], exchange_name=exchange_name)
        if equities:
            return equities[0]
        return None
        # async with self.session_maker() as session:
        #     q_equity_symbol_mapping = select(EquitySymbolMappingModel).where(
        #         EquitySymbolMappingModel.exchange == exchange_name,
        #         EquitySymbolMappingModel.symbol == symbol)
        #     equity_mapping = (await session.execute(q_equity_symbol_mapping)).scalar_one_or_none()
        #     if equity_mapping is None:
        #         return None
        #     q_equity = select(EquityModel).where(EquityModel.sid == equity_mapping.sid).options(
        #         selectinload(EquityModel.asset_router)).options(selectinload(EquityModel.equity_symbol_mappings))
        #     asset: EquityModel = (await session.execute(q_equity)).scalar_one_or_none()
        #
        #     if asset is None:
        #         return None
        #     return Equity(
        #         sid=asset.sid,
        #         asset_name=asset.asset_name,
        #         start_date=asset.start_date,
        #         first_traded=asset.first_traded,
        #         end_date=asset.end_date,
        #         auto_close_date=asset.auto_close_date,
        #         symbol_mapping={
        #             equity_mapping.exchange: EquitySymbolMapping(
        #                 company_symbol=equity_mapping.company_symbol,
        #                 symbol=equity_mapping.symbol,
        #                 exchange_name=equity_mapping.exchange,
        #                 share_class_symbol=equity_mapping.share_class_symbol,
        #                 end_date=equity_mapping.end_date,
        #                 start_date=equity_mapping.start_date
        #             )
        #             for equity_mapping in asset.equity_symbol_mappings
        #         }
        #     )

    def migrate(self) -> None:
        alembic_dir_path = Path(pathlib.Path(__file__).parent.parent.parent, "alembic")
        alembic_cfg = config.Config(Path(alembic_dir_path, "alembic.ini"))
        alembic_cfg.set_main_option("script_location", str(Path(alembic_dir_path)))
        alembic_cfg.set_main_option("sqlalchemy.url", self.db_url.replace("+aiosqlite", ""))
        # os.makedirs(db_path.parent, exist_ok=True)
        # Run the migration
        command.upgrade(alembic_cfg, "head")

    @property
    def exchange_info(self):
        with self.engine.connect() as conn:
            es = conn.execute(sa.select(self.exchanges.c)).fetchall()
        return {
            name: ExchangeInfo(name, canonical_name, country_code)
            for name, canonical_name, country_code in es
        }

    @property
    def symbol_ownership_map(self):
        out = {}
        for mappings in self.symbol_ownership_maps_by_country_code.values():
            for key, ownership_periods in mappings.items():
                out.setdefault(key, []).extend(ownership_periods)

        return out

    @property
    def symbol_ownership_maps_by_country_code(self):
        with self.engine.connect() as conn:
            query = sa.select(
                self.equities.c.sid,
                self.exchanges.c.country_code,
            ).where(self.equities.c.exchange == self.exchanges.c.exchange)
            sid_to_country_code = dict(conn.execute(query).fetchall())

            return build_grouped_ownership_map(
                conn,
                table=self.equity_symbol_mappings,
                key_from_row=(lambda row: (row.company_symbol, row.share_class_symbol)),
                value_from_row=lambda row: row.symbol,
                group_key=lambda row: sid_to_country_code[row.sid],
            )

    def lookup_asset_types(self, sids: list[int]):
        """Retrieve asset types for a list of sids.

        Parameters
        ----------
        sids : list[int]

        Returns
        -------
        types : dict[sid -> str or None]
            Asset types for the provided sids.
        """
        found = {}
        missing = set()

        for sid in sids:
            try:
                found[sid] = self._asset_type_cache[sid]
            except KeyError:
                missing.add(sid)

        if not missing:
            return found

        router_cols = self.asset_router.c

        with self.engine.connect() as conn:
            for assets in group_into_chunks(missing):
                query = sa.select(router_cols.sid, router_cols.asset_type).where(
                    self.asset_router.c.sid.in_(map(int, assets))
                )
                for sid, type_ in conn.execute(query).fetchall():
                    missing.remove(sid)
                    found[sid] = self._asset_type_cache[sid] = type_

                for sid in missing:
                    found[sid] = self._asset_type_cache[sid] = None

        return found

    def group_by_type(self, sids: list[int]):
        """Group a list of sids by asset type.

        Parameters
        ----------
        sids : list[int]

        Returns
        -------
        types : dict[str or None -> list[int]]
            A dict mapping unique asset types to lists of sids drawn from sids.
            If we fail to look up an asset, we assign it a key of None.
        """
        return invert(self.lookup_asset_types(sids))

    def retrieve_asset(self, sid: int, default_none: bool = False):
        """
        Retrieve the Asset for a given sid.
        """
        try:
            asset = self._asset_cache[sid]
            if asset is None and not default_none:
                raise SidsNotFound(sids=[sid])
            return asset
        except KeyError:
            return self.retrieve_all(sids=[sid, ], default_none=default_none)[0]

    async def retrieve_all(self, sids: list[int], default_none: bool = False):
        """Retrieve all assets in `sids`.

        Parameters
        ----------
        sids : iterable of int
            Assets to retrieve.
        default_none : bool
            If True, return None for failed lookups.
            If False, raise `SidsNotFound`.

        Returns
        -------
        assets : list[AssetModel or None]
            A list of the same length as `sids` containing Assets (or Nones)
            corresponding to the requested sids.

        Raises
        ------
        SidsNotFound
            When a requested sid is not found and default_none=False.
        """

        async with self.session_maker() as session:
            q = select(AssetModel).where(AssetModel.sid.in_(sids))
            assets = (await session.execute(q)).scalars()
            return list(assets)

        hits, missing, failures = {}, set(), []
        for sid in sids:
            try:
                asset = self._asset_cache[sid]
                if not default_none and asset is None:
                    # Bail early if we've already cached that we don't know
                    # about an asset.
                    raise SidsNotFound(sids=[sid])
                hits[sid] = asset
            except KeyError:
                missing.add(sid)

        # All requests were cache hits.  Return requested sids in order.
        if not missing:
            return [hits[sid] for sid in sids]

        update_hits = hits.update

        # Look up cache misses by type.
        type_to_assets = self.group_by_type(sids=missing)

        # Handle failures
        failures = {failure: None for failure in type_to_assets.pop(None, ())}
        update_hits(failures)
        self._asset_cache.update(failures)

        if failures and not default_none:
            raise SidsNotFound(sids=list(failures))

        # We don't update the asset cache here because it should already be
        # updated by `self.retrieve_equities`.
        update_hits(self.retrieve_equities(sids=type_to_assets.pop("equity", [])))
        update_hits(self.retrieve_futures_contracts(sids=type_to_assets.pop("future", [])))

        # We shouldn't know about any other asset types.
        if type_to_assets:
            raise AssertionError("Found asset types: %s" % list(type_to_assets.keys()))

        return [hits[sid] for sid in sids]

    def retrieve_equities(self, sids: list[int]):
        """Retrieve Equity objects for a list of sids.

        Users generally shouldn't need to this method (instead, they should
        prefer the more general/friendly `retrieve_assets`), but it has a
        documented interface and tests because it's used upstream.

        Parameters
        ----------
        sids : iterable[int]

        Returns
        -------
        equities : dict[int -> Equity]

        Raises
        ------
        EquitiesNotFound
            When any requested asset isn't found.
        """
        return self._retrieve_assets(sids=sids, asset_tbl=self.equities, asset_type=Equity)

    # def _retrieve_equity(self, sid):
    #     return self.retrieve_equities(sids=[sid, ])[sid]

    def retrieve_futures_contracts(self, sids: list[int]):
        """Retrieve Future objects for an iterable of sids.

        Users generally shouldn't need to this method (instead, they should
        prefer the more general/friendly `retrieve_assets`), but it has a
        documented interface and tests because it's used upstream.

        Parameters
        ----------
        sids : iterable[int]

        Returns
        -------
        equities : dict[int -> Equity]

        Raises
        ------
        EquitiesNotFound
            When any requested asset isn't found.
        """
        return self._retrieve_assets(sids=sids, asset_tbl=self.futures_contracts, asset_type=FuturesContract)

    @staticmethod
    def _select_assets_by_sid(asset_tbl: Table, sids: list[int]):
        return sa.select(asset_tbl).where(asset_tbl.c.sid.in_(map(int, sids)))

    @staticmethod
    def _select_asset_by_symbol(asset_tbl: Table, symbol: str):
        return sa.select(asset_tbl).where(asset_tbl.c.symbol == symbol)

    def _select_most_recent_symbols_chunk(self, sid_group: list[int]):
        """Retrieve the most recent symbol for a set of sids.

        Parameters
        ----------
        sid_group : iterable[int]
            The sids to lookup. The length of this sequence must be less than
            or equal to SQLITE_MAX_VARIABLE_NUMBER because the sids will be
            passed in as sql bind params.

        Returns
        -------
        sel : Selectable
            The sqlalchemy selectable that will query for the most recent
            symbol for each sid.

        Notes
        -----
        This is implemented as an inner select of the columns of interest
        ordered by the end date of the (sid, symbol) mapping. We then group
        that inner select on the sid with no aggregations to select the last
        row per group which gives us the most recently active symbol for all
        of the sids.
        """
        cols = self.equity_symbol_mappings.c

        # These are the columns we actually want.
        data_cols = (cols.sid,) + tuple(cols[name] for name in SYMBOL_COLUMNS)

        # Also select the max of end_date so that all non-grouped fields take
        # on the value associated with the max end_date.
        # to_select = data_cols + (sa.func.max(cols.end_date),)
        func_rank = (
            sa.func.rank()
            .over(order_by=cols.end_date.desc(), partition_by=cols.sid)
            .label("rnk")
        )
        to_select = data_cols + (func_rank,)

        subquery = (
            sa.select(*to_select)
            .where(cols.sid.in_(map(int, sid_group)))
            .subquery("sq")
        )
        query = (
            sa.select(subquery.columns)
            .filter(subquery.c.rnk == 1)
            .select_from(subquery)
        )
        return query

    def _lookup_most_recent_symbols(self, sids: list[int]):
        with self.engine.connect() as conn:
            return {
                row.sid: {c: row[c] for c in SYMBOL_COLUMNS}
                for row in concat(
                    conn.execute(self._select_most_recent_symbols_chunk(sid_group=sid_group))
                    .mappings()
                    .fetchall()
                    for sid_group in partition_all(n=SQLITE_MAX_VARIABLE_NUMBER, seq=sids)
                )
            }

    def _retrieve_asset_dicts(self, sids: list[int], asset_tbl: Table, querying_equities):
        if not sids:
            return

        if querying_equities:

            def mkdict(
                    row,
                    exchanges=self.exchange_info,
                    symbols=self._lookup_most_recent_symbols(sids=sids),
            ):
                d = dict(row)
                d["exchange_info"] = exchanges[d.pop("exchange")]
                # we are not required to have a symbol for every asset, if
                # we don't have any symbols we will just use the empty string
                return merge(d, symbols.get(row["sid"], {}))

        else:

            def mkdict(row, exchanges=self.exchange_info):
                d = dict(row)
                d["exchange_info"] = exchanges[d.pop("exchange")]
                return d

        for assets in group_into_chunks(sids):
            # Load misses from the db.
            query = self._select_assets_by_sid(asset_tbl, assets)

            with self.engine.connect() as conn:
                for row in conn.execute(query).mappings().fetchall():
                    yield _convert_asset_timestamp_fields(mkdict(row))

    def _retrieve_assets(self, sids: list[int], asset_tbl: Table, asset_type: type):
        """Internal function for loading assets from a table.

        This should be the only method of `AssetFinder` that writes Assets into
        self._asset_cache.

        Parameters
        ---------
        sids : iterable of int
            Asset ids to look up.
        asset_tbl : sqlalchemy.Table
            Table from which to query assets.
        asset_type : type
            Type of asset to be constructed.

        Returns
        -------
        assets : dict[int -> Asset]
            Dict mapping requested sids to the retrieved assets.
        """
        # Fastpath for empty request.
        if not sids:
            return {}

        cache = self._asset_cache
        hits = {}

        querying_equities = issubclass(asset_type, Equity)
        filter_kwargs = (
            _filter_equity_kwargs if querying_equities else _filter_future_kwargs
        )

        rows = self._retrieve_asset_dicts(sids, asset_tbl, querying_equities)
        for row in rows:
            sid = row["sid"]
            asset = asset_type(**filter_kwargs(row))
            hits[sid] = cache[sid] = asset

        # If we get here, it means something in our code thought that a
        # particular sid was an equity/future and called this function with a
        # concrete type, but we couldn't actually resolve the asset.  This is
        # an error in our code, not a user-input error.
        misses = tuple(set(sids) - hits.keys())
        if misses:
            if querying_equities:
                raise EquitiesNotFound(sids=misses)
            else:
                raise FutureContractsNotFound(sids=misses)
        return hits

    def _lookup_symbol_strict(self, ownership_map: dict[(str, str), list[OwnershipPeriod]], multi_country: bool,
                              symbol: str, as_of_date: datetime.datetime):
        """Resolve a symbol to an asset object without fuzzy matching.

        Parameters
        ----------
        ownership_map : dict[(str, str), list[OwnershipPeriod]]
            The mapping from split symbols to ownership periods.
        multi_country : bool
            Does this mapping span multiple countries?
        symbol : str
            The symbol to look up.
        as_of_date : datetime or None
            If multiple assets have held this sid, which day should the
            resolution be checked against? If this value is None and multiple
            sids have held the ticker, then a MultipleSymbolsFound error will
            be raised.

        Returns
        -------
        asset : AssetModel
            The asset that held the given symbol.

        Raises
        ------
        SymbolNotFound
            Raised when the symbol or symbol as_of_date pair do not map to
            any assets.
        MultipleSymbolsFound
            Raised when multiple assets held the symbol. This happens if
            multiple assets held the symbol at disjoint times and
            ``as_of_date`` is None, or if multiple assets held the symbol at
            the same time and``multi_country`` is True.

        Notes
        -----
        The resolution algorithm is as follows:

        - Split the symbol into the company and share class component.
        - Do a dictionary lookup of the
          ``(company_symbol, share_class_symbol)`` in the provided ownership
          map.
        - If there is no entry in the dictionary, we don't know about this
          symbol so raise a ``SymbolNotFound`` error.
        - If ``as_of_date`` is None:
          - If more there is more than one owner, raise
            ``MultipleSymbolsFound``
          - Otherwise, because the list mapped to a symbol cannot be empty,
            return the single asset.
        - Iterate through all of the owners:
          - If the ``as_of_date`` is between the start and end of the ownership
            period:
            - If multi_country is False, return the found asset.
            - Otherwise, put the asset in a list.
        - At the end of the loop, if there are no candidate assets, raise a
          ``SymbolNotFound``.
        - If there is exactly one candidate, return it.
        - Othewise, raise ``MultipleSymbolsFound`` because the ticker is not
          unique across countries.
        """
        # split the symbol into the components, if there are no
        # company/share class parts then share_class_symbol will be empty
        company_symbol, share_class_symbol = split_delimited_symbol(symbol=symbol)
        try:
            owners = ownership_map[company_symbol, share_class_symbol]
            assert owners, "empty owners list for %r" % symbol
        except KeyError as exc:
            # no equity has ever held this symbol
            raise SymbolNotFound(symbol=symbol) from exc

        if not as_of_date:
            # exactly one equity has ever held this symbol, we may resolve
            # without the date
            if len(owners) == 1:
                return self.retrieve_asset(sid=owners[0].sid)

            options = {self.retrieve_asset(sid=owner.sid) for owner in owners}

            if multi_country:
                country_codes = map(attrgetter("country_code"), options)

                if len(set(country_codes)) > 1:
                    raise SameSymbolUsedAcrossCountries(
                        symbol=symbol, options=dict(zip(country_codes, options))
                    )

            # more than one equity has held this ticker, this
            # is ambiguous without the date
            raise MultipleSymbolsFound(symbol=symbol, options=options)

        options = []
        country_codes = []
        for start, end, sid, _ in owners:
            if start.date() <= as_of_date < end.date():
                # find the equity that owned it on the given asof date
                asset = self.retrieve_asset(sid=sid)

                # if this asset owned the symbol on this asof date and we are
                # only searching one country, return that asset
                if not multi_country:
                    return asset
                else:
                    options.append(asset)
                    country_codes.append(asset.country_code)

        if not options:
            # no equity held the ticker on the given asof date
            raise SymbolNotFound(symbol=symbol)

        # if there is one valid option given the asof date, return that option
        if len(options) == 1:
            return options[0]

        # if there's more than one option given the asof date, a country code
        # must be passed to resolve the symbol to an asset
        raise SameSymbolUsedAcrossCountries(
            symbol=symbol, options=dict(zip(country_codes, options))
        )

    def _choose_symbol_ownership_map(self, country_code: str):
        if country_code is None:
            return self.symbol_ownership_map

        return self.symbol_ownership_maps_by_country_code.get(country_code)

    def lookup_symbol(self, symbol: str, as_of_date: datetime.datetime,
                      country_code: str | None = None):
        """Lookup an equity by symbol.

        Parameters
        ----------
        symbol : str
            The ticker symbol to resolve.
        as_of_date : datetime.datetime or None
            Look up the last owner of this symbol as of this datetime.
            If ``as_of_date`` is None, then this can only resolve the equity
            if exactly one equity has ever owned the ticker.
        country_code : str or None, optional
            The country to limit searches to. If not provided, the search will
            span all countries which increases the likelihood of an ambiguous
            lookup.

        Returns
        -------
        equity : Equity
            The equity that held ``symbol`` on the given ``as_of_date``, or the
            only equity to hold ``symbol`` if ``as_of_date`` is None.

        Raises
        ------
        SymbolNotFound
            Raised when no equity has ever held the given symbol.
        MultipleSymbolsFound
            Raised when no ``as_of_date`` is given and more than one equity
            has held ``symbol``. This is also raised when ``fuzzy=True`` and
            there are multiple candidates for the given ``symbol`` on the
            ``as_of_date``. Also raised when no ``country_code`` is given and
            the symbol is ambiguous across multiple countries.
        """
        if symbol is None:
            raise TypeError(
                "Cannot lookup asset for symbol of None for "
                "as of date %s." % as_of_date
            )

        f = self._lookup_symbol_strict
        mapping = self._choose_symbol_ownership_map(country_code)

        if mapping is None:
            raise SymbolNotFound(symbol=symbol)
        return f(
            mapping,
            country_code is None,
            symbol,
            as_of_date,
        )

    # def lookup_symbols(self, symbols: list[str], as_of_date: datetime.datetime,
    #                    country_code: str | None = None):
    #     """Lookup a list of equities by symbol.
    #
    #     Equivalent to::
    #
    #         [finder.lookup_symbol(s, as_of, fuzzy) for s in symbols]
    #
    #     but potentially faster because repeated lookups are memoized.
    #
    #     Parameters
    #     ----------
    #     symbols : sequence[str]
    #         Sequence of ticker symbols to resolve.
    #     as_of_date : datetime.datetime
    #         Forwarded to ``lookup_symbol``.
    #     country_code : str or None, optional
    #         The country to limit searches to. If not provided, the search will
    #         span all countries which increases the likelihood of an ambiguous
    #         lookup.
    #
    #     Returns
    #     -------
    #     equities : list[Equity]
    #     """
    #     if not symbols:
    #         return []
    #
    #     multi_country = country_code is None
    #     f = self._lookup_symbol_strict
    #     mapping = self._choose_symbol_ownership_map(country_code)
    #
    #     if mapping is None:
    #         raise SymbolNotFound(symbol=symbols[0])
    #
    #     memo = {}
    #     out = []
    #     append_output = out.append
    #     for sym in symbols:
    #         if sym in memo:
    #             append_output(memo[sym])
    #         else:
    #             equity = memo[sym] = f(
    #                 mapping,
    #                 multi_country,
    #                 sym,
    #                 as_of_date,
    #             )
    #             append_output(equity)
    #     return out

    def lookup_future_symbol(self, symbol: str):
        """Lookup a future contract by symbol.

        Parameters
        ----------
        symbol : str
            The symbol of the desired contract.

        Returns
        -------
        future : Future
            The future contract referenced by ``symbol``.

        Raises
        ------
        SymbolNotFound
            Raised when no contract named 'symbol' is found.

        """
        with self.engine.connect() as conn:
            data = (
                conn.execute(
                    self._select_asset_by_symbol(asset_tbl=self.futures_contracts, symbol=symbol)
                )
                .mappings()
                .fetchone()
            )

        # If no data found, raise an exception
        if not data:
            raise SymbolNotFound(symbol=symbol)
        return self.retrieve_asset(sid=data["sid"])

    # def lookup_by_supplementary_field(self, field_name: str, value: float, as_of_date: datetime.datetime):
    #     try:
    #         owners = self.equity_supplementary_map[
    #             field_name,
    #             value,
    #         ]
    #         assert owners, "empty owners list for field %r (sid: %r)" % (
    #             field_name,
    #             value,
    #         )
    #     except KeyError as exc:
    #         # no equity has ever held this value
    #         raise ValueNotFoundForField(field=field_name, value=value) from exc
    #
    #     if not as_of_date:
    #         if len(owners) > 1:
    #             # more than one equity has held this value, this is ambigious
    #             # without the date
    #             raise MultipleValuesFoundForField(
    #                 field=field_name,
    #                 value=value,
    #                 options=set(
    #                     map(
    #                         compose(self.retrieve_asset, attrgetter("sid")),
    #                         owners,
    #                     )
    #                 ),
    #             )
    #         # exactly one equity has ever held this value, we may resolve
    #         # without the date
    #         return self.retrieve_asset(owners[0].sid)
    #
    #     for start, end, sid, _ in owners:
    #         if start <= as_of_date < end:
    #             # find the equity that owned it on the given asof date
    #             return self.retrieve_asset(sid)
    #
    #     # no equity held the value on the given asof date
    #     raise ValueNotFoundForField(field=field_name, value=value)

    # def get_supplementary_field(self, sid: int, field_name: str, as_of_date: datetime.datetime):
    #     """Get the value of a supplementary field for an asset.
    #
    #     Parameters
    #     ----------
    #     sid : int
    #         The sid of the asset to query.
    #     field_name : str
    #         Name of the supplementary field.
    #     as_of_date : datetime.datetime, None
    #         The last known value on this date is returned. If None, a
    #         value is returned only if we've only ever had one value for
    #         this sid. If None and we've had multiple values,
    #         MultipleValuesFoundForSid is raised.
    #
    #     Raises
    #     ------
    #     NoValueForSid
    #         If we have no values for this asset, or no values was known
    #         on this as_of_date.
    #     MultipleValuesFoundForSid
    #         If we have had multiple values for this asset over time, and
    #         None was passed for as_of_date.
    #     """
    #     try:
    #         periods = self.equity_supplementary_map_by_sid[
    #             field_name,
    #             sid,
    #         ]
    #         assert periods, "empty periods list for field %r and sid %r" % (
    #             field_name,
    #             sid,
    #         )
    #     except KeyError:
    #         raise NoValueForSid(field=field_name, sid=sid) from KeyError
    #
    #     if not as_of_date:
    #         if len(periods) > 1:
    #             # This equity has held more than one value, this is ambigious
    #             # without the date
    #             raise MultipleValuesFoundForSid(
    #                 field=field_name,
    #                 sid=sid,
    #                 options={p.value for p in periods},
    #             )
    #         # this equity has only ever held this value, we may resolve
    #         # without the date
    #         return periods[0].value
    #
    #     for start, end, _, value in periods:
    #         if start <= as_of_date < end:
    #             return value
    #
    #     # Could not find a value for this sid on the as_of_date.
    #     raise NoValueForSid(field=field_name, sid=sid)

    def _get_contract_sids(self, root_symbol: str):
        fc_cols = self.futures_contracts.c
        with self.engine.connect() as conn:
            return (
                conn.execute(
                    sa.select(
                        fc_cols.sid,
                    )
                    .where(
                        (fc_cols.root_symbol == root_symbol)
                        & (fc_cols.start_date != pd.NaT.value)
                    )
                    .order_by(fc_cols.sid)
                )
                .scalars()
                .fetchall()
            )

    def _get_root_symbol_exchange(self, root_symbol: str):
        fc_cols = self.futures_root_symbols.c
        fields = (fc_cols.exchange,)

        with self.engine.connect() as conn:
            exchange = conn.execute(
                sa.select(*fields).where(fc_cols.root_symbol == root_symbol)
            ).scalar()

        if exchange is not None:
            return exchange
        else:
            raise SymbolNotFound(symbol=root_symbol)

    def get_ordered_contracts(self, root_symbol: str):
        try:
            return self._ordered_contracts[root_symbol]
        except KeyError:
            contract_sids = self._get_contract_sids(root_symbol)
            contracts = deque(self.retrieve_all(contract_sids))
            chain_predicate = self._future_chain_predicates.get(root_symbol, None)
            oc = OrderedContracts(root_symbol, contracts, chain_predicate)
            self._ordered_contracts[root_symbol] = oc
            return oc

    def create_continuous_future(self, root_symbol: str, offset: int, roll_style: str, adjustment: str):
        if adjustment not in ADJUSTMENT_STYLES:
            raise ValueError(
                f"Invalid adjustment style {adjustment!r}. Allowed adjustment styles are "
                f"{list(ADJUSTMENT_STYLES)}."
            )

        oc = self.get_ordered_contracts(root_symbol=root_symbol)
        exchange = self._get_root_symbol_exchange(root_symbol=root_symbol)

        sid = _encode_continuous_future_sid(root_symbol=root_symbol, offset=offset, roll_style=roll_style,
                                            adjustment_style=None)
        mul_sid = _encode_continuous_future_sid(root_symbol=root_symbol, offset=offset, roll_style=roll_style,
                                                adjustment_style="div")
        add_sid = _encode_continuous_future_sid(root_symbol=root_symbol, offset=offset, roll_style=roll_style,
                                                adjustment_style="add")

        cf_template = partial(
            ContinuousFuture,
            root_symbol=root_symbol,
            offset=offset,
            roll_style=roll_style,
            start_date=oc.start_date,
            end_date=oc.end_date,
            exchange_info=self.exchange_info[exchange],
        )

        cf = cf_template(sid=sid)
        mul_cf = cf_template(sid=mul_sid, adjustment="mul")
        add_cf = cf_template(sid=add_sid, adjustment="add")

        self._asset_cache[cf.sid] = cf
        self._asset_cache[mul_cf.sid] = mul_cf
        self._asset_cache[add_cf.sid] = add_cf

        return {None: cf, "mul": mul_cf, "add": add_cf}[adjustment]

    #
    # def _get_sids(self, tblattr: str) -> list[int]:
    #     with self.engine.connect() as conn:
    #         return list((
    #             conn.execute(sa.select(getattr(self, tblattr).c.sid))
    #             .scalars()
    #             .fetchall()
    #         ))

    # @property
    # def sids(self) -> list[int]:
    #     return self._get_sids("asset_router")

    def _lookup_generic_scalar(self, obj: AssetModel, as_of_date: datetime.datetime, country_code: str,
                               matches: list[AssetModel],
                               missing: list[AssetModel]):
        """
        Convert asset_convertible to an asset.

        On success, append to matches.
        On failure, append to missing.
        """
        result = self._lookup_generic_scalar_helper(
            obj=obj,
            as_of_date=as_of_date,
            country_code=country_code,
        )
        if result is not None:
            matches.append(result)
        else:
            missing.append(obj)

    def _lookup_generic_scalar_helper(self, obj: AssetModel, as_of_date: datetime.datetime, country_code: str):
        if isinstance(obj, (AssetModel, ContinuousFuture)):
            return obj

        if isinstance(obj, Integral):
            try:
                return self.retrieve_asset(int(obj))
            except SidsNotFound:
                return None

        if isinstance(obj, str):
            # Try to look up as an equity first.
            try:
                return self.lookup_symbol(
                    symbol=obj, as_of_date=as_of_date, country_code=country_code
                )
            except SymbolNotFound:
                # Fall back to lookup as a Future
                try:
                    # TODO: Support country_code for future_symbols?
                    return self.lookup_future_symbol(obj)
                except SymbolNotFound:
                    return None

        raise NotAssetConvertible("Input was %s, not AssetConvertible." % obj)

    def _compute_asset_lifetimes(self, **kwargs):
        """Compute and cache a recarray of asset lifetimes"""
        sids = starts = ends = []
        equities_cols = self.equities.c
        exchanges_cols = self.exchanges.c
        if len(kwargs) == 1:
            if "country_codes" in kwargs.keys():
                condt = exchanges_cols.country_code.in_(kwargs["country_codes"])
            if "exchange_names" in kwargs.keys():
                condt = exchanges_cols.exchange.in_(kwargs["exchange_names"])

            with self.engine.connect() as conn:
                results = conn.execute(
                    sa.select(
                        equities_cols.sid,
                        equities_cols.start_date,
                        equities_cols.end_date,
                    ).where(
                        (exchanges_cols.exchange == equities_cols.exchange) & (condt)
                    )
                ).fetchall()
            if results:
                sids, starts, ends = zip(*results)

        sid = np.array(sids, dtype="i8")
        start = np.array(starts, dtype="f8")
        end = np.array(ends, dtype="f8")
        start[np.isnan(start)] = 0  # convert missing starts to 0
        end[np.isnan(end)] = np.iinfo(int).max  # convert missing end to INTMAX
        return Lifetimes(sid, start.astype("i8"), end.astype("i8"))

    def lifetimes(self, dates: pd.DatetimeIndex, include_start_date: bool, country_codes: list[str]):
        """Compute a DataFrame representing asset lifetimes for the specified date
        range.

        Parameters
        ----------
        dates : pd.DatetimeIndex
            The dates for which to compute lifetimes.
        include_start_date : bool
            Whether or not to count the asset as alive on its start_date.

            This is useful in a backtesting context where `lifetimes` is being
            used to signify "do I have data for this asset as of the morning of
            this date?"  For many financial metrics, (e.g. daily close), data
            isn't available for an asset until the end of the asset's first
            day.
        country_codes : iterable[str]
            The country codes to get lifetimes for.

        Returns
        -------
        lifetimes : pd.DataFrame
            A frame of dtype bool with `dates` as index and an Int64Index of
            assets as columns.  The value at `lifetimes.loc[date, asset]` will
            be True iff `asset` existed on `date`.  If `include_start_date` is
            False, then lifetimes.loc[date, asset] will be false when date ==
            asset.start_date.

        See Also
        --------
        numpy.putmask
        ziplime.pipeline.engine.SimplePipelineEngine._compute_root_mask
        """
        if isinstance(country_codes, str):
            raise TypeError(
                "Got string {!r} instead of an iterable of strings in "
                "AssetFinder.lifetimes.".format(country_codes),
            )

        # normalize to a cache-key so that we can memoize results.
        country_codes = frozenset(country_codes)

        lifetimes = self._asset_lifetimes.get(country_codes)
        if lifetimes is None:
            self._asset_lifetimes[country_codes] = lifetimes = (
                self._compute_asset_lifetimes(country_codes=country_codes)
            )

        raw_dates = as_column(dates.asi8)
        if include_start_date:
            mask = lifetimes.start <= raw_dates
        else:
            mask = lifetimes.start < raw_dates
        mask &= raw_dates <= lifetimes.end

        return pd.DataFrame(mask, index=dates, columns=lifetimes.sid)

    # def equities_sids_for_country_code(self, country_code: str):
    #     """Return all of the sids for a given country.
    #
    #     Parameters
    #     ----------
    #     country_code : str
    #         An ISO 3166 alpha-2 country code.
    #
    #     Returns
    #     -------
    #     tuple[int]
    #         The sids whose exchanges are in this country.
    #     """
    #     sids = self._compute_asset_lifetimes(country_codes=[country_code]).sid
    #     return tuple(sids.tolist())

    # def equities_sids_for_exchange_name(self, exchange_name: str):
    #     """Return all of the sids for a given exchange_name.
    #
    #     Parameters
    #     ----------
    #     exchange_name : str
    #
    #     Returns
    #     -------
    #     tuple[int]
    #         The sids whose exchanges are in this country.
    #     """
    #     sids = self._compute_asset_lifetimes(exchange_names=[exchange_name]).sid
    #     return tuple(sids.tolist())

    def to_json(self):
        return {
            "db_url": self.db_url
        }

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Self:
        return cls(
            db_url=data["db_url"],
            future_chain_predicates=CHAIN_PREDICATES
        )



from math import sqrt

import numpy as np
import pandas as pd
import structlog

from ziplime.assets.entities.futures_contract import FuturesContract

logger = structlog.get_logger(__name__)


class PositionStats:
    """Computed values from the current positions.

    Attributes
    ----------
    gross_exposure : float64
        The gross position exposure.
    gross_value : float64
        The gross position value.
    long_exposure : float64
        The exposure of just the long positions.
    long_value : float64
        The value of just the long positions.
    net_exposure : float64
        The net position exposure.
    net_value : float64
        The net position value.
    short_exposure : float64
        The exposure of just the short positions.
    short_value : float64
        The value of just the short positions.
    longs_count : int64
        The number of long positions.
    shorts_count : int64
        The number of short positions.
    position_exposure_array : np.ndarray[float64]
        The exposure of each position in the same order as
        ``position_tracker.positions``.
    position_exposure_series : pd.Series[float64]
        The exposure of each position in the same order as
        ``position_tracker.positions``. The index is the numeric sid of each
        asset.

    Notes
    -----
    ``position_exposure_array`` and ``position_exposure_series`` share the same
    underlying memory. The array interface should be preferred if you are doing
    access each minute for better performance.

    ``position_exposure_array`` and ``position_exposure_series`` may be mutated
    when the position tracker next updates the stats. Do not rely on these
    objects being preserved across accesses to ``stats``. If you need to freeze
    the values, you must take a copy.
    """

    # cdef readonly np.float64_t gross_exposure
    # cdef readonly np.float64_t gross_value
    # cdef readonly np.float64_t long_exposure
    # cdef readonly np.float64_t long_value
    # cdef readonly np.float64_t net_exposure
    # cdef readonly np.float64_t net_value
    # cdef readonly np.float64_t short_exposure
    # cdef readonly np.float64_t short_value
    # cdef readonly np.uint64_t longs_count
    # cdef readonly np.uint64_t shorts_count
    # cdef readonly object position_exposure_array
    # cdef readonly object position_exposure_series
    #
    # # These are the same memory exposed through ``position_exposure_array``
    # # and ``position_exposure_series``. These are hidden from Python.
    # cdef object underlying_value_array
    # cdef object underlying_index_array

    def __init__(self):
        self.gross_exposure = 0.0
        self.gross_value = 0.0
        self.long_exposure = 0.0
        self.long_value = 0.0
        self.net_exposure = 0.0
        self.net_value = 0.0
        self.short_exposure = 0.0
        self.short_value = 0.0
        self.longs_count = 0
        self.shorts_count = 0
        self.position_exposure_array = None
        self.position_exposure_series = None

        # These are the same memory exposed through ``position_exposure_array``
        # and ``position_exposure_series``. These are hidden from Python.
        self.underlying_value_array = None
        self.underlying_index_array = None

    @classmethod
    def new(cls):
        self = cls()
        self.position_exposure_series = es = pd.Series(
            np.array([], dtype='float64'),
            index=np.array([], dtype='int64'),
        )
        self.underlying_value_array = self.position_exposure_array = es.values
        self.underlying_index_array = es.index.values
        return self


def calculate_position_tracker_stats(positions, stats):
    """Calculate various stats about the current positions.

    Parameters
    ----------
    positions : OrderedDict
        The ordered dictionary of positions.

    Returns
    -------
    position_stats : PositionStats
        The computed statistics.
    """
    npos = len(positions)

    old_index = stats.underlying_index_array
    old_position_exposure = (
        stats.underlying_value_array
    )

    long_value = 0.0
    short_value = 0.0

    long_exposure = 0.0
    short_exposure = 0.0

    longs_count = 0
    shorts_count = 0

    # attempt to reuse the memory of the old exposure series
    if len(old_index) < npos:
        # we don't have enough space in the cached buffer, allocate a new
        # array
        stats.underlying_index_array = index = np.empty(npos, dtype='int64')
        stats.underlying_value_array = position_exposure = np.empty(
            npos,
            dtype='float64',
        )

        stats.position_exposure_array = position_exposure
        # create a new series to expose the arrays
        stats.position_exposure_series = pd.Series(
            position_exposure,
            index=index,
        )
    elif len(old_index) > npos:
        # we have more space than needed, slice off the extra but leave it
        # available
        index = old_index[:npos]
        position_exposure = old_position_exposure[:npos]

        stats.position_exposure_array = position_exposure
        # create a new series with the sliced arrays
        stats.position_exposure_series = pd.Series(
            position_exposure,
            index=index,
        )
    else:
        # we have exactly the right amount of space, no slicing or allocation
        # needed
        index = old_index
        position_exposure = old_position_exposure

        stats.position_exposure_array = position_exposure
        stats.position_exposure_series = pd.Series(
            position_exposure,
            index=index,
        )

    ix = 0

    for outer_position in positions.values():
        position = outer_position

        # NOTE: this loop does a lot of stuff!
        # we call this function every time the portfolio value is needed,
        # which is at least once per simulation day, so let's not iterate
        # through every single position multiple times.
        # try:
        exposure = position.amount * position.last_sale_price
        # except Exception as e:
        #     print("exception multiplying a")
        #     raise
        if type(position.asset) is FuturesContract:
            # Futures don't have an inherent position value.
            value = 0

            # unchecked cast, this is safe because we do a type check above
            exposure *= position.asset.price_multiplier
        else:
            value = exposure

        if exposure > 0:
            longs_count += 1
            long_value += value
            long_exposure += exposure
        elif exposure < 0:
            shorts_count += 1
            short_value += value
            short_exposure += exposure

        index[ix] = position.asset.sid
        position_exposure[ix] = exposure

        ix += 1

    net_value = long_value + short_value
    gross_value = long_value - short_value

    net_exposure = long_exposure + short_exposure
    gross_exposure = long_exposure - short_exposure

    stats.gross_exposure = gross_exposure
    stats.gross_value = gross_value
    stats.long_exposure = long_exposure
    stats.long_value = long_value
    stats.longs_count = longs_count
    stats.net_exposure = net_exposure
    stats.net_value = net_value
    stats.short_exposure = short_exposure
    stats.short_value = short_value
    stats.shorts_count = shorts_count


def minute_annual_volatility(date_labels,
                             minute_returns,
                             daily_returns):
    """Pre-compute the minute cumulative volatility field.
    """
    out = np.empty_like(minute_returns)
    previous_date = date_labels[0]
    day_ix = 0
    daily_sum = 0
    todays_prod = 1
    annualization_factor = sqrt(252.0)

    for ix in range(len(minute_returns)):
        date = date_labels[ix]
        this_minute_returns = minute_returns[ix]

        if date != previous_date:
            previous_date = date
            daily_sum += daily_returns[day_ix]
            day_ix += 1
            todays_prod = 1

        if day_ix < 1:
            variance = np.nan
        else:
            todays_prod *= 1 + this_minute_returns

            intermediate_sum = daily_sum + todays_prod - 1
            mean = intermediate_sum / (day_ix + 1)

            variance = todays_prod - 1 - mean
            variance *= variance  # squared

            demeaned_old = daily_returns[:day_ix] - mean
            variance += demeaned_old.dot(demeaned_old)

            # variance /= day_ix  # day_count - 1 for ddof=1
            variance = variance/day_ix
        # print(ix, sqrt(variance) * annualization_factor)
        res = sqrt(variance) * annualization_factor
        if np.isnan(res):
            out[ix] = 0
        else:
            out[ix] = res

    return out



#
# Copyright 2015 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain

import pandas as pd
from numpy import (
    int64,
    uint32,
    zeros,
)
from pandas import Timestamp

from ziplime.lib.adjustment import Float64Multiply

from ziplime.utils.pandas_utils import timedelta_to_integral_seconds
from ziplime.utils.sqlite_utils import SQLITE_MAX_VARIABLE_NUMBER

_SID_QUERY_TEMPLATE = """
SELECT DISTINCT sid FROM {0}
WHERE effective_date >= ? AND effective_date <= ?
"""
SID_QUERIES = {
    tablename: _SID_QUERY_TEMPLATE.format(tablename)
    for tablename in ('splits', 'dividends', 'mergers')
}

ADJ_QUERY_TEMPLATE = """
SELECT sid, ratio, effective_date
FROM {0}
WHERE sid IN ({1}) AND effective_date >= {2} AND effective_date <= {3}
"""

EPOCH = Timestamp(0, tz='UTC')


def _get_sids_from_table(db,
                         tablename: str,
                         start_date: int,
                         end_date: int) -> set:
    """Get the unique sids for all adjustments between start_date and end_date
    from table `tablename`.

    Parameters
    ----------
    db : sqlite3.connection
    tablename : str
    start_date : int (seconds since epoch)
    end_date : int (seconds since epoch)

    Returns
    -------
    sids : set
        Set of sets
    """

    cursor = db.execute(
        SID_QUERIES[tablename],
        (start_date, end_date),
    )
    out = set()
    for result in cursor.fetchall():
        out.add(result[0])
    return out


def _get_split_sids(db, start_date: int, end_date: int) -> set:
    return _get_sids_from_table(db, 'splits', start_date, end_date)


def _get_merger_sids(db, start_date: int, end_date: int) -> set:
    return _get_sids_from_table(db, 'mergers', start_date, end_date)


def _get_dividend_sids(db, start_date: int, end_date: int) -> set:
    return _get_sids_from_table(db, 'dividends', start_date, end_date)


def _adjustments(adjustments_db,
                 split_sids: set,
                 merger_sids: set,
                 dividends_sids: set,
                 start_date: int,
                 end_date: int,
                 assets: pd.Index):
    c = adjustments_db.cursor()

    splits_to_query = [str(a) for a in assets if a in split_sids]
    splits_results = []
    while splits_to_query:
        query_len = min(len(splits_to_query), SQLITE_MAX_VARIABLE_NUMBER)
        query_assets = splits_to_query[:query_len]
        t = [str(a) for a in query_assets]
        statement = ADJ_QUERY_TEMPLATE.format(
            'splits',
            ",".join(['?' for _ in query_assets]),
            start_date,
            end_date,
        )
        c.execute(statement, t)
        splits_to_query = splits_to_query[query_len:]
        splits_results.extend(c.fetchall())

    mergers_to_query = [str(a) for a in assets if a in merger_sids]
    mergers_results = []
    while mergers_to_query:
        query_len = min(len(mergers_to_query), SQLITE_MAX_VARIABLE_NUMBER)
        query_assets = mergers_to_query[:query_len]
        t = [str(a) for a in query_assets]
        statement = ADJ_QUERY_TEMPLATE.format(
            'mergers',
            ",".join(['?' for _ in query_assets]),
            start_date,
            end_date,
        )
        c.execute(statement, t)
        mergers_to_query = mergers_to_query[query_len:]
        mergers_results.extend(c.fetchall())

    dividends_to_query = [str(a) for a in assets if a in dividends_sids]
    dividends_results = []
    while dividends_to_query:
        query_len = min(len(dividends_to_query), SQLITE_MAX_VARIABLE_NUMBER)
        query_assets = dividends_to_query[:query_len]
        t = [str(a) for a in query_assets]
        statement = ADJ_QUERY_TEMPLATE.format(
            'dividends',
            ",".join(['?' for _ in query_assets]),
            start_date,
            end_date,
        )
        c.execute(statement, t)
        dividends_to_query = dividends_to_query[query_len:]
        dividends_results.extend(c.fetchall())

    return splits_results, mergers_results, dividends_results


def load_adjustments_from_sqlite(adjustments_db,
                                 dates: pd.DatetimeIndex,
                                 assets: pd.Index,
                                 should_include_splits: bool,
                                 should_include_mergers: bool,
                                 should_include_dividends: bool,
                                 adjustment_type: str):
    """Load a dictionary of Adjustment objects from adjustments_db.

    Parameters
    ----------
    adjustments_db : sqlite3.Connection
        Connection to a sqlite3 table in the format written by
        SQLiteAdjustmentWriter.
    dates : pd.DatetimeIndex
        Dates for which adjustments are needed.
    assets : pd.Int64Index
        Assets for which adjustments are needed.
    should_include_splits : bool
        Whether split adjustments should be included.
    should_include_mergers : bool
        Whether merger adjustments should be included.
    should_include_dividends : bool
        Whether dividend adjustments should be included.
    adjustment_type : str
        Whether price adjustments, volume adjustments, or both, should be
        included in the output.

    Returns
    -------
    adjustments : dict[str -> dict[int -> Adjustment]]
        A dictionary containing price and/or volume adjustment mappings from
        index to adjustment objects to apply at that index.
    """

    if not (adjustment_type == 'price' or
            adjustment_type == 'volume' or
            adjustment_type == 'all'):
        raise ValueError(
            "%s is not a valid adjustment type.\n"
            "Valid adjustment types are 'price', 'volume', and 'all'.\n" % (
                adjustment_type,
            )
        )

    should_include_price_adjustments = bool(
        adjustment_type == 'all' or adjustment_type == 'price'
    )
    should_include_volume_adjustments = bool(
        adjustment_type == 'all' or adjustment_type == 'volume'
    )

    if not should_include_price_adjustments:
        should_include_mergers = False
        should_include_dividends = False

    start_date = timedelta_to_integral_seconds(dates[0] - EPOCH)
    end_date = timedelta_to_integral_seconds(dates[-1] - EPOCH)

    if should_include_splits:
        split_sids = _get_split_sids(
            adjustments_db,
            start_date,
            end_date,
        )
    else:
        split_sids = set()

    if should_include_mergers:
        merger_sids = _get_merger_sids(
            adjustments_db,
            start_date,
            end_date,
        )
    else:
        merger_sids = set()

    if should_include_dividends:
        dividend_sids = _get_dividend_sids(
            adjustments_db,
            start_date,
            end_date,
        )
    else:
        dividend_sids = set()

    splits, mergers, dividends = _adjustments(
        adjustments_db,
        split_sids,
        merger_sids,
        dividend_sids,
        start_date,
        end_date,
        assets,
    )

    price_adjustments = {}
    volume_adjustments = {}
    result = {}
    asset_ixs = {}  # Cache sid lookups here.
    date_ixs = {}

    _dates_seconds = \
        dates.values.astype('datetime64[s]').view(int64)

    # Pre-populate date index cache.
    for i, dt in enumerate(_dates_seconds):
        date_ixs[dt] = i

    # splits affect prices and volumes, volumes is the inverse
    for sid, ratio, eff_date in splits:
        if eff_date < start_date:
            continue

        date_loc = _lookup_dt(date_ixs, eff_date, _dates_seconds)

        if sid not in asset_ixs:
            asset_ixs[sid] = assets.get_loc(sid)
        asset_ix = asset_ixs[sid]

        if should_include_price_adjustments:
            price_adj = Float64Multiply(0, date_loc, asset_ix, asset_ix, ratio)
            price_adjustments.setdefault(date_loc, []).append(price_adj)

        if should_include_volume_adjustments:
            volume_adj = Float64Multiply(
                0, date_loc, asset_ix, asset_ix, 1.0 / ratio
            )
            volume_adjustments.setdefault(date_loc, []).append(volume_adj)

    # mergers and dividends affect prices only
    for sid, ratio, eff_date in chain(mergers, dividends):
        if eff_date < start_date:
            continue

        date_loc = _lookup_dt(date_ixs, eff_date, _dates_seconds)

        if sid not in asset_ixs:
            asset_ixs[sid] = assets.get_loc(sid)
        asset_ix = asset_ixs[sid]

        price_adj = Float64Multiply(0, date_loc, asset_ix, asset_ix, ratio)
        price_adjustments.setdefault(date_loc, []).append(price_adj)

    if should_include_price_adjustments:
        result['price'] = price_adjustments
    if should_include_volume_adjustments:
        result['volume'] = volume_adjustments

    return result


def _lookup_dt(dt_cache: dict,
               dt: int,
               fallback):
    if dt not in dt_cache:
        dt_cache[dt] = fallback.searchsorted(dt, side='right')
    return dt_cache[dt]



def bulleted_list(items, max_count=None, indent=2):
    """Format a bulleted list of values."""
    if max_count is not None and len(items) > max_count:
        item_list = list(items)
        items = item_list[: max_count - 1]
        items.append("...")
        items.append(item_list[-1])

    line_template = (" " * indent) + "- {}"
    return "\n".join(map(line_template.format, items))



import datetime

import asyncclick as click



class _DatetimeParam(click.ParamType):
    def __init__(self, tz=None):
        self.tz = tz

    def parser(self, value):
        return value
        return datetime.datetime.strptime(value, "%Y-%m-%d").replace(tzinfo=self.tz)

    @property
    def name(self):
        return type(self).__name__.upper()

    def convert(self, value, param, ctx):
        try:
            return self.parser(value)
        except ValueError:
            self.fail(
                "%s is not a valid %s" % (value, self.name.lower()),
                param,
                ctx,
            )


class Timestamp(_DatetimeParam):
    """A click parameter that parses the value into pandas.Timestamp objects.

    Parameters
    ----------
    tz : timezone-coercable, optional
        The timezone to parse the string as.
        By default the timezone will be infered from the string or naiive.
    """


class Date(_DatetimeParam):
    """A click parameter that parses the value into datetime.date objects.

    Parameters
    ----------
    tz : timezone-coercable, optional
        The timezone to parse the string as.
        By default the timezone will be infered from the string or naiive.
    as_timestamp : bool, optional
        If True, return the value as a pd.Timestamp object normalized to
        midnight.
    """

    def __init__(self, tz=None, as_timestamp=False):
        super(Date, self).__init__(tz=tz)
        self.as_timestamp = as_timestamp

    def parser(self, value):
        ts = super(Date, self).parser(value)
        return ts.normalize() if self.as_timestamp else ts.date()




import datetime

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.slippage.slippage_model import SlippageModel


class FixedSlippage(SlippageModel):
    """Simple model assuming a fixed-size spread for all assets.

    Parameters
    ----------
    spread : float, optional
        Size of the assumed spread for all assets.
        Orders to buy will be filled at ``close + (spread / 2)``.
        Orders to sell will be filled at ``close - (spread / 2)``.

    Notes
    -----
    This model does not impose limits on the size of fills. An order for an
    asset will always be filled as soon as any trading activity occurs in the
    order's asset, even if the size of the order is greater than the historical
    volume.
    """

    def __init__(self, spread=0.0):
        super(FixedSlippage, self).__init__()
        self.spread = spread

    def __repr__(self):
        return "{class_name}(spread={spread})".format(
            class_name=self.__class__.__name__,
            spread=self.spread,
        )

    def process_order(self, exchange: Exchange, dt:datetime.datetime, order):
        price = exchange.current(order.asset, "close")

        return (price + (self.spread / 2.0 * order.direction), order.amount)



import datetime

from ziplime.errors import LiquidityExceeded
from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.order import Order
from ziplime.finance.slippage.slippage_model import SlippageModel


class FixedBasisPointsSlippage(SlippageModel):
    """
    Model slippage as a fixed percentage difference from historical minutely
    close price, limiting the size of fills to a fixed percentage of historical
    minutely volume.

    Orders to buy are filled at::

        historical_price * (1 + (basis_points * 0.0001))

    Orders to sell are filled at::

        historical_price * (1 - (basis_points * 0.0001))

    Fill sizes are capped at::

        historical_volume * volume_limit

    Parameters
    ----------
    basis_points : float, optional
        Number of basis points of slippage to apply for each fill. Default
        is 5 basis points.
    volume_limit : float, optional
        Fraction of trading volume that can be filled each minute. Default is
        10% of trading volume.

    Notes
    -----
    - A basis point is one one-hundredth of a percent.
    - This class, default-constructed, is ziplime's default slippage model for
      equities.
    """

    def __init__(self, basis_points=5.0, volume_limit=0.1):
        super(FixedBasisPointsSlippage, self).__init__()
        if volume_limit <= 0:
            raise ValueError("volume_limit must be positive.")
        if basis_points <= 0:
            raise ValueError("volume_limit must be positive.")

        self.basis_points = basis_points
        self.percentage = float(self.basis_points) / 10000.0
        self.volume_limit = volume_limit

    def __repr__(self):
        return """
{class_name}(
    basis_points={basis_points},
    volume_limit={volume_limit},
)
""".strip().format(
            class_name=self.__class__.__name__,
            basis_points=self.basis_points,
            volume_limit=self.volume_limit,
        )

    def process_order(self, exchange: Exchange, dt:datetime.datetime, order: Order) -> tuple[float, float]:
        current_val = exchange.current(assets=frozenset({order.asset}), fields=frozenset({"close", "volume"}), dt=dt)
        volume = current_val["volume"][0]
        max_volume = int(self.volume_limit * volume)

        price = current_val["close"][0]
        shares_to_fill = min(abs(order.open_amount), max_volume - self.volume_for_bar)

        if shares_to_fill == 0:
            raise LiquidityExceeded()

        return (
            price + price * (self.percentage * order.direction),
            shares_to_fill * order.direction,
        )



from collections import defaultdict, OrderedDict
from copy import copy

import structlog

from .blotter import Blotter

from ziplime.domain.bar_data import BarData
from ziplime.finance.domain.order import Order
from ziplime.exchanges.exchange import Exchange
from ...assets.entities.asset import Asset


class InMemoryBlotter(Blotter):
    def __init__(
            self,
            exchanges: dict[str, Exchange],
            cancel_policy,
    ):
        super().__init__(cancel_policy=cancel_policy)
        self._logger = structlog.get_logger(__name__)
        self.exchanges = exchanges
        # these orders are aggregated by asset
        self.open_orders = {exchange.name: defaultdict(dict) for exchange in exchanges.values()}
        # keep a dict of orders by their own id
        self.orders = {exchange.name: {} for exchange in exchanges.values()}
        # holding orders that have come in since the last event.
        self.new_orders = OrderedDict()

        self.max_shares = int(1e11)

    def save_order(self, order: Order):
        """Place an order.

        Parameters
        ----------
        asset : ziplime.assets.Asset
            The asset that this order is for.
        amount : int
            The amount of shares to order. If ``amount`` is positive, this is
            the number of shares to buy or cover. If ``amount`` is negative,
            this is the number of shares to sell or short.
        style : ziplime.finance.execution.ExecutionStyle
            The execution style for the order.
        order_id : str, optional
            The unique identifier for this order.

        Returns
        -------
        order_id : str or None
            The unique identifier for this order, or None if no order was
            placed.

        Notes
        -----
        amount > 0 :: Buy/Cover
        amount < 0 :: Sell/Short
        Market order:    order(asset, amount)
        Limit order:     order(asset, amount, style=LimitOrder(limit_price))
        Stop order:      order(asset, amount, style=StopOrder(stop_price))
        StopLimit order: order(asset, amount, style=StopLimitOrder(limit_price,
        stop_price))
        """
        # something could be done with amount to further divide
        # between buy by share count OR buy shares up to a dollar amount
        # numeric == share count  AND  "$dollar.cents" == cost amount
        self.open_orders[order.exchange_name][order.asset][order.id] = order
        self.orders[order.exchange_name][order.id] = order
        return order.id

    def order_cancelled(self, order: Order) -> None:
        asset_orders = self.open_orders[order.exchange_name][order.asset.sid]
        asset_orders.pop(order.id, None)

    def order_rejected(self, order: Order) -> None:
        asset_orders = self.open_orders[order.exchange_name][order.asset.sid]
        asset_orders.pop(order.id, None)

    def get_order_by_id(self, order_id: str, exchange_name: str) -> Order | None:
        return self.orders.get(exchange_name, {}).get(order_id, None)

    def get_open_orders_by_asset(self, asset: Asset, exchange_name: str) -> dict[str, Order] | None:
        return self.open_orders.get(exchange_name, {}).get(asset, None)

    def get_open_orders(self, exchange_name: str) -> dict[Asset, dict[str, Order]]:
        return self.open_orders.get(exchange_name, {})

    def get_all_assets_in_open_orders(self) -> list[Asset]:
        return [asset for asset_orders in self.open_orders.values() for asset in asset_orders]

    def cancel_all_orders_for_asset(self, asset: Asset, exchange_name: str, relay_status: bool = True):
        """
        Cancel all open orders for a given asset.
        """
        self.open_orders.get(exchange_name, {}).pop(asset, None)

    # End of day cancel for daily frequency
    def execute_daily_cancel_policy(self, event):
        if self.cancel_policy.should_cancel(event):
            warn = self.cancel_policy.warn_on_cancel
            for asset in copy(self.open_orders):
                orders = self.open_orders[asset]
                if len(orders) > 1:
                    order = orders[0]
                    self.cancel(order.id, relay_status=True)
                    if warn:
                        if order.filled > 0:
                            self._logger.warn(
                                f"Your order for {order.amount} shares of "
                                f"{order.asset.symbol} has been partially filled. "
                                f"{order.filled} shares were successfully "
                                f"purchased. {order.amount - order.filled} shares were not "
                                f"filled by the end of day and "
                                f"were canceled."
                            )
                        elif order.filled < 0:
                            self._logger.warn(
                                f"Your order for {order.amount} shares of "
                                f"{order.asset.symbol} has been partially filled. "
                                f"{-1 * order.filled} shares were successfully "
                                f"sold. {-1 * (order.amount - order.filled)} shares were not "
                                f"filled by the end of day and "
                                f"were canceled."
                            )
                        else:
                            self._logger.warn(
                                f"Your order for {order.amount} shares of "
                                f"{order.asset.symbol} failed to fill by the end of day "
                                f"and was canceled."
                            )

    def execute_cancel_policy(self, event):
        if self.cancel_policy.should_cancel(event):
            warn = self.cancel_policy.warn_on_cancel
            for exchange in self.open_orders:
                for asset in self.open_orders[exchange]:
                    self.cancel_all_orders_for_asset(asset=asset, exchange_name=exchange, relay_status=False)

    def order_held(self, order: Order) -> None:
        pass

    def process_splits(self, splits):
        """
        Processes a list of splits by modifying any open orders as needed.

        Parameters
        ----------
        splits: list
            A list of splits.  Each split is a tuple of (asset, ratio).

        Returns
        -------
        None
        """
        for asset, ratio in splits:
            if asset not in self.open_orders:
                continue

            orders_to_modify = self.open_orders[asset]
            for order in orders_to_modify:
                order.handle_split(ratio)

    # def get_transactions(self, bar_data: BarData):
    #     """
    #     Creates a list of transactions based on the current open orders,
    #     slippage model, and commission model.
    #
    #     Parameters
    #     ----------
    #     bar_data: ziplime.protocol.BarData
    #
    #     Notes
    #     -----
    #     This method book-keeps the blotter's open_orders dictionary, so that
    #      it is accurate by the time we're done processing open orders.
    #
    #     Returns
    #     -------
    #     transactions_list: List
    #         transactions_list: list of transactions resulting from the current
    #         open orders.  If there were no open orders, an empty list is
    #         returned.
    #
    #     commissions_list: List
    #         commissions_list: list of commissions resulting from filling the
    #         open orders.  A commission is an object with "asset" and "cost"
    #         parameters.
    #
    #     closed_orders: List
    #         closed_orders: list of all the orders that have filled.
    #     """
    #
    #     closed_orders = []
    #     transactions = []
    #     commissions = []
    #
    #     if self.open_orders:
    #         for asset, asset_orders in self.open_orders.items():
    #             slippage = self.exchange.get_slippage_model(asset=asset)
    #
    #             for order, txn in slippage.simulate(data=bar_data, assets=frozenset({asset}),
    #                                                 orders_for_asset=asset_orders.values()):
    #                 commission = self.exchange.get_commission_model(asset=asset)
    #                 additional_commission = commission.calculate(order, txn)
    #
    #                 if additional_commission > 0:
    #                     commissions.append(
    #                         {
    #                             "asset": order.asset,
    #                             "order": order,
    #                             "cost": additional_commission,
    #                         }
    #                     )
    #
    #                 order.filled += txn.amount
    #                 order.commission += additional_commission
    #
    #                 order.dt = txn.dt
    #
    #                 transactions.append(txn)
    #
    #                 if not order.open:
    #                     closed_orders.append(order)
    #
    #     return transactions, commissions, closed_orders

    def prune_orders(self, closed_orders):
        """
        Removes all given orders from the blotter's open_orders list.

        Parameters
        ----------
        closed_orders: iterable of orders that are closed.

        Returns
        -------
        None
        """
        # remove all closed orders from our open_orders dict
        for order in closed_orders:
            asset = order.asset
            asset_orders = self.open_orders[order.exchange_name][asset]
            asset_orders.pop(order, None)

        # now clear out the assets from our open_orders dict that have
        # zero open orders
        for exchange_name, exchange_orders in self.open_orders.items():
            for asset in list(exchange_orders.keys()):
                if len(self.open_orders[exchange_name][asset]) == 0:
                    self.open_orders.get(exchange_name, {}).pop(asset, None)



#
# Copyright 2016 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc

from abc import abstractmethod

from ziplime.trading.enums.simulation_event import SimulationEvent


class CancelPolicy(metaclass=abc.ABCMeta):
    """Abstract cancellation policy interface."""

    @abstractmethod
    def should_cancel(self, event):
        """Should all open orders be cancelled?

        Parameters
        ----------
        event : enum-value
            An event type, one of:
              - :data:`ziplime.gens.sim_engine.BAR`
              - :data:`ziplime.gens.sim_engine.DAY_START`
              - :data:`ziplime.gens.sim_engine.DAY_END`
              - :data:`ziplime.gens.sim_engine.MINUTE_END`

        Returns
        -------
        should_cancel : bool
            Should all open orders be cancelled?
        """
        pass


class EODCancel(CancelPolicy):
    """This policy cancels open orders at the end of the day.  For now,
    Zipline will only apply this policy to minutely simulations.

    Parameters
    ----------
    warn_on_cancel : bool, optional
        Should a warning be raised if this causes an order to be cancelled?
    """

    def __init__(self, warn_on_cancel=True):
        self.warn_on_cancel = warn_on_cancel

    def should_cancel(self, event):
        return event == SimulationEvent.SESSION_END


class NeverCancel(CancelPolicy):
    """Orders are never automatically canceled."""

    def __init__(self):
        self.warn_on_cancel = False

    def should_cancel(self, event):
        return False



from .base import FXRateReader


class ExplodingFXRateReader(FXRateReader):
    """An FXRateReader that raises an error when used.

    This is useful for testing contexts where FX rates aren't actually needed.
    """

    def get_rates(self, rate, quote, bases, dts):
        raise AssertionError("FX rates requested unexpectedly!")



import asyncio
import datetime
from pathlib import Path

from ziplime.utils.calendar_utils import get_calendar

from ziplime.assets.domain.ordered_contracts import CHAIN_PREDICATES
from ziplime.assets.repositories.sqlalchemy_adjustments_repository import SqlAlchemyAdjustmentRepository
from ziplime.assets.repositories.sqlalchemy_asset_repository import SqlAlchemyAssetRepository
from ziplime.assets.services.asset_service import AssetService
from ziplime.constants.data_type import DataType
from ziplime.core.algorithm_file import AlgorithmFile
from ziplime.data.services.bundle_service import BundleService
from ziplime.data.services.file_system_bundle_registry import FileSystemBundleRegistry
from ziplime.exchanges.lime_trader_sdk.lime_trader_sdk_exchange import LimeTraderSdkExchange
from ziplime.finance.metrics import default_metrics
from ziplime.gens.domain.realtime_clock import RealtimeClock
from ziplime.exchanges.exchange import Exchange
from ziplime.utils.run_algo import run_algorithm


async def _run_live_trading(
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        trading_calendar: str,
        emission_rate: datetime.timedelta,
        cash_balance: float,
        bundle_name: str,
        algorithm_file: str,
        exchange: Exchange = None,
        config_file: str | None = None,

):
    # benchmark_spec = BenchmarkSpec(
    #     benchmark_returns=None,
    #     benchmark_sid=benchmark_sid,
    #     benchmark_symbol=benchmark_symbol,
    #     benchmark_file=benchmark_file,
    #     no_benchmark=no_benchmark,
    # )
    calendar = get_calendar(trading_calendar)

    bundle_storage_path = str(Path(Path.home(), ".ziplime", "data"))
    bundle_registry = FileSystemBundleRegistry(base_data_path=bundle_storage_path)
    bundle_service = BundleService(bundle_registry=bundle_registry)
    data_bundle = None
    if bundle_name is not None:
        data_bundle = await bundle_service.load_bundle(bundle_name=bundle_name, bundle_version=None,
                                                       data_type=DataType.CUSTOM)

    algo = AlgorithmFile(algorithm_file=algorithm_file, algorithm_config_file=config_file)
    timedelta_diff_from_current_time = -datetime.timedelta(seconds=0)

    clock = RealtimeClock(
        trading_calendar=calendar,
        start_date=start_date,
        end_date=end_date,
        emission_rate=emission_rate,
        timedelta_diff_from_current_time=timedelta_diff_from_current_time
    )

    if exchange is None:
        exchange = LimeTraderSdkExchange(
            name="LIME",
            country_code="US",
            trading_calendar=calendar,
            data_bundle=data_bundle,
            cash_balance=cash_balance,
            clock=clock
        )

    db_url = f"sqlite+aiosqlite:///{str(Path(Path.home(), ".ziplime", "assets.sqlite").absolute())}"
    assets_repository = SqlAlchemyAssetRepository(db_url=db_url, future_chain_predicates=CHAIN_PREDICATES)
    adjustments_repository = SqlAlchemyAdjustmentRepository(db_url=db_url)
    asset_service = AssetService(asset_repository=assets_repository, adjustments_repository=adjustments_repository)

    return await run_algorithm(
        algorithm=algo,
        asset_service=asset_service,
        print_algo=True,
        metrics_set=default_metrics(),
        # benchmark_spec=benchmark_spec,
        custom_loader=None,
        exchanges=[exchange],
        clock=clock,
    )


def run_live_trading(
        trading_calendar: str,
        algorithm_file: str,
        total_cash: float,
        emission_rate: datetime.timedelta,
        start_date: datetime.datetime | None = None,
        end_date: datetime.datetime | None = None,
        bundle_name: str | None = None,
        config_file: str | None = None,
        exchange: Exchange | None = None
):
    return asyncio.run(_run_live_trading(start_date=start_date, end_date=end_date, trading_calendar=trading_calendar,
                                         cash_balance=total_cash,
                                         algorithm_file=algorithm_file,
                                         config_file=config_file, bundle_name=bundle_name, exchange=exchange,
                                         emission_rate=emission_rate))



import datetime
import enum


class DataFrequency(enum.Enum):
    SECOND = "1s"
    MINUTE = "1m"
    MINUTE_3 = "3m"
    MINUTE_5 = "5m"
    MINUTE_15 = "15m"
    MINUTE_30 = "30m"

    HOUR = "1h"
    HOUR_2 = "1h"
    HOUR_4 = "4h"
    HOUR_6 = "6h"
    HOUR_8 = "8h"
    HOUR_12 = "12h"

    DAY = "1d"
    DAY_3 = "1d"

    WEEK = "1w"

    MONTH = "1M"

    def to_timedelta(self) -> datetime.timedelta:
        time_unit, value = self.value[-1:], int(self.value[:-1])
        match time_unit:
            case "s":
                return datetime.timedelta(seconds=value)
            case "m":
                return datetime.timedelta(minutes=value)
            case "h":
                return datetime.timedelta(hours=value)
            case "d":
                return datetime.timedelta(days=value)
            case "w":
                return datetime.timedelta(weeks=value)
            case "M":
                return datetime.timedelta(days=30)
            case _:
                raise ValueError(f"Invalid time unit: {time_unit}")



import datetime
import polars as pl
from typing import Any, Self

from ziplime.assets.entities.asset import Asset
from ziplime.assets.models.dividend import Dividend


class AdjustmentRepository:
    def get_splits(self, assets: frozenset[Asset], dt: datetime.date): ...

    def get_stock_dividends(self, sid: int, trading_days: pl.Series) -> list[Dividend]: ...

    def to_json(self): ...

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Self: ...



def s(word, seq, suffix="s"):
    """Adds a suffix to ``word`` if some sequence has anything other than
    exactly one element.

    Parameters
    ----------
    word : str
        The string to add the suffix to.
    seq : sequence
        The sequence to check the length of.
    suffix : str, optional.
        The suffix to add to ``word``

    Returns
    -------
    maybe_plural : str
        ``word`` with ``suffix`` added if ``len(seq) != 1``.
    """
    if len(seq) == 1:
        return word

    return word + suffix


def plural(singular, plural, seq):
    """Selects a singular or plural word based on the length of a sequence.

    Parameters
    ----------
    singlular : str
        The string to use when ``len(seq) == 1``.
    plural : str
        The string to use when ``len(seq) != 1``.
    seq : sequence
        The sequence to check the length of.

    Returns
    -------
    maybe_plural : str
        Either ``singlular`` or ``plural``.
    """
    if len(seq) == 1:
        return singular

    return plural


def bulleted_list(items, indent=0, bullet_type="-"):
    """Format a bulleted list of values.

    Parameters
    ----------
    items : sequence
        The items to make a list.
    indent : int, optional
        The number of spaces to add before each bullet.
    bullet_type : str, optional
        The bullet type to use.

    Returns
    -------
    formatted_list : str
        The formatted list as a single string.
    """
    format_string = " " * indent + bullet_type + " {}"
    return "\n".join(map(format_string.format, items))



import datetime

from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import DateIndexed, IntegerIndexed, StringPK
from ziplime.core.db.base_model import BaseModel


class StockDividendPayout(BaseModel):
    __tablename__ = "stock_dividend_payouts"

    index: Mapped[StringPK]
    sid: Mapped[IntegerIndexed]
    ex_date: Mapped[DateIndexed]

    declared_date: Mapped[datetime.date]
    record_date: Mapped[datetime.date]
    pay_date: Mapped[datetime.date]
    payment_sid: Mapped[str]
    ration: Mapped[float]



from functools import wraps

import ziplime.api
from ziplime.utils.algo_instance import get_algo_instance, set_algo_instance


class ZiplineAPI(object):
    """
    Context manager for making an algorithm instance available to ziplime API
    functions within a scoped block.
    """

    def __init__(self, algo_instance):
        self.algo_instance = algo_instance

    def __enter__(self):
        """
        Set the given algo instance, storing any previously-existing instance.
        """
        self.old_algo_instance = get_algo_instance()
        set_algo_instance(self.algo_instance)

    def __exit__(self, _type, _value, _tb):
        """
        Restore the algo instance stored in __enter__.
        """
        set_algo_instance(self.old_algo_instance)


def api_method(f):
    # Decorator that adds the decorated class method as a callable
    # function (wrapped) to ziplime.api
    @wraps(f)
    def wrapped(*args, **kwargs):
        # Get the instance and call the method
        algo_instance = get_algo_instance()
        if algo_instance is None:
            raise RuntimeError(
                'ziplime api method %s must be called during a simulation.'
                % f.__name__
            )
        return getattr(algo_instance, f.__name__)(*args, **kwargs)
    # Add functor to ziplime.api
    setattr(ziplime.api, f.__name__, wrapped)
    ziplime.api.__all__.append(f.__name__)
    f.is_api_method = True
    return f


def require_not_initialized(exception):
    """
    Decorator for API methods that should only be called during or before
    TradingAlgorithm.initialize.  `exception` will be raised if the method is
    called after initialize.

    Usage
    -----
    @require_not_initialized(SomeException("Don't do that!"))
    def method(self):
        # Do stuff that should only be allowed during initialize.
    """
    def decorator(method):
        @wraps(method)
        def wrapped_method(self, *args, **kwargs):
            if self.initialized:
                raise exception
            return method(self, *args, **kwargs)
        return wrapped_method
    return decorator


def require_initialized(exception):
    """
    Decorator for API methods that should only be called after
    TradingAlgorithm.initialize.  `exception` will be raised if the method is
    called before initialize has completed.

    Usage
    -----
    @require_initialized(SomeException("Don't do that!"))
    def method(self):
        # Do stuff that should only be allowed after initialize.
    """
    def decorator(method):
        @wraps(method)
        def wrapped_method(self, *args, **kwargs):
            if not self.initialized:
                raise exception
            return method(self, *args, **kwargs)
        return wrapped_method
    return decorator


def disallowed_in_before_trading_start(exception):
    """
    Decorator for API methods that cannot be called from within
    TradingAlgorithm.before_trading_start.  `exception` will be raised if the
    method is called inside `before_trading_start`.

    Usage
    -----
    @disallowed_in_before_trading_start(SomeException("Don't do that!"))
    def method(self):
        # Do stuff that is not allowed inside before_trading_start.
    """
    def decorator(method):
        @wraps(method)
        def wrapped_method(self, *args, **kwargs):
            if self._in_before_trading_start:
                raise exception
            return method(self, *args, **kwargs)
        return wrapped_method
    return decorator


# def allowed_only_in_before_trading_start(exception):
#     """
#     Decorator for API methods that can be called only from within
#     TradingAlgorithm.before_trading_start.  `exception` will be raised if the
#     method is called outside `before_trading_start`.
#
#     Usage
#     -----
#     @allowed_only_in_before_trading_start(SomeException("Don't do that!"))
#     def method(self):
#         # Do stuff that is only allowed inside before_trading_start.
#     """
#     def decorator(method):
#         @wraps(method)
#         def wrapped_method(self, *args, **kwargs):
#             if not self._in_before_trading_start:
#                 raise exception
#             return method(self, *args, **kwargs)
#         return wrapped_method
#     return decorator



import polars as pl
import datetime

import structlog
from exchange_calendars import ExchangeCalendar

from ziplime.assets.services.asset_service import AssetService
from ziplime.constants.period import Period

_logger = structlog.get_logger(__name__)
async def _process_data(data: pl.DataFrame,
                        date_start: datetime.datetime,
                        date_end: datetime.datetime,
                        data_frequency_use_window_end: bool,
                        asset_service: AssetService, name: str, symbols: list[str],
                        trading_calendar: ExchangeCalendar,
                        frequency: datetime.timedelta | Period, ):
    """Ingest data for a given bundle.        """
    _logger.info(f"Ingesting custom bundle: name={name}, date_start={date_start}, date_end={date_end}, "
                      f"symbols={symbols}, frequency={frequency}")
    if date_start < trading_calendar.first_session.replace(tzinfo=trading_calendar.tz):
        raise ValueError(
            f"Date start must be after first session of trading calendar. "
            f"First session is {trading_calendar.first_session.replace(tzinfo=trading_calendar.tz)} "
            f"and date start is {date_start}")

    if date_end > trading_calendar.last_session.replace(tzinfo=trading_calendar.tz):
        raise ValueError(
            f"Date end must be before last session of trading calendar. "
            f"Last session is {trading_calendar.last_session.replace(tzinfo=trading_calendar.tz)} "
            f"and date end is {date_end}")

    if data.is_empty():
        _logger.warning(
            f"No data for symbols={symbols}, frequency={frequency}, date_start={date_start},"
            f"date_end={date_end} found. Skipping ingestion."
        )
        return

    # repair data
    all_bars = [
        s for s in pl.from_pandas(
            trading_calendar.sessions_minutes(start=date_start.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None),
                                              end=date_end.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None)).tz_convert(trading_calendar.tz)
        ) if s >= date_start and s <= date_end
    ]

    required_sessions = pl.DataFrame({"date": all_bars}).group_by_dynamic(
        index_column="date", every=frequency
    ).agg()
    if data_frequency_use_window_end:
        if (
                (type(frequency) is datetime.timedelta and frequency >= datetime.timedelta(days=1)) or
                (type(frequency) is str and frequency in ["1d", "1w", "1mo", "1q", "1y"])
        ):
            last_row = required_sessions.tail(1).with_columns(
                pl.col("date").dt.offset_by(frequency) - pl.duration(days=1))
            required_sessions = required_sessions.with_columns(
                pl.col("date") - pl.duration(days=1)
            )[1:]

            required_sessions = pl.concat([required_sessions, last_row])
    required_columns = [
        "date"
    ]
    missing = [c for c in required_columns if c not in data.columns]

    if missing:
        raise ValueError(f"Ingested data is missing required columns: {missing}. Cannot ingest bundle.")
    if "symbol" not in data.columns and "sid" not in data.columns:
        raise ValueError(f"When ingesting custom bundle you must supply either a symbol or a sid column.")

    sid_id = "sid" in data.columns
    symbol_id = "symbol" in data.columns

    asset_identifiers = list(data["sid"].unique()) if sid_id else list(data["symbol"].unique())

    if sid_id:
        data = await _backfill_symbol_data(data=data, asset_service=asset_service,
                                                required_sessions=required_sessions)
    else:
        data = await _backfill_sid_data(data=data, asset_service=asset_service,
                                             required_sessions=required_sessions)
    return data


async def _backfill_sid_data(data: pl.DataFrame, asset_service: AssetService, required_sessions: pl.Series):
    unique_symbols = list(data["symbol"].unique())
    symbol_to_sid = {a.get_symbol_by_exchange(exchange_name=None): a.sid for a in
                     await asset_service.get_equities_by_symbols(unique_symbols)}
    data = data.with_columns(
        pl.lit(0).alias("sid"),
    )

    for symbol in unique_symbols:
        symbol_data = data.filter(symbol=symbol).with_columns(pl.col("date"))
        missing_sessions = sorted(set(required_sessions["date"]) - set(symbol_data["date"]))

        if len(missing_sessions) > 0:
            _logger.warning(
                f"Data for symbol {symbol} is missing on ticks ({len(missing_sessions)}): {[missing_session.isoformat() for missing_session in missing_sessions]}")
            new_rows_df = pl.DataFrame(
                {"date": missing_sessions, "symbol": symbol},
                schema_overrides={"date": data.schema["date"]}
            )
            # Concatenate with the original DataFrame
            data = pl.concat([data, new_rows_df], how="diagonal")

        data = data.with_columns(
            pl.col("symbol").replace(symbol_to_sid).cast(pl.Int64).alias("sid")
        ).sort(["sid", "date"])
    return data


async def _backfill_symbol_data(self):
    pass



import enum
from typing import Any

import numpy as np
import pandas as pd
from pandas import isnull, Timestamp
from numpy import asarray, datetime64, float64, int64, bool_, uint8

from ziplime.utils.compat import unicode


# cimport numpy as np
#
# # Purely for readability. There aren't C-level declarations for these types.
# ctypedef object Int64Index_t
# ctypedef object DatetimeIndex_t
# ctypedef object Timestamp_t

#
# ctypedef fused column_type:
#     np.int64_t
#     np.float64_t
#     np.uint8_t
#     object
def _from_assets_and_dates(cls,
                           dates_index: pd.DatetimeIndex,
                           assets_index: pd.RangeIndex,
                           start_date: pd.Timestamp,
                           end_date: pd.Timestamp,
                           asset_id: int,
                           value):
    """
    Helper for constructing an Adjustment instance from coordinates in
    assets/dates indices.

    Example
    -------

    >>> from pandas import date_range,  Timestamp
    >>> dates = date_range('2014-01-01', '2014-01-07')
    >>> assets = Int64Index(range(10))
    >>> Float64Multiply.from_assets_and_dates(
    ...     dates,
    ...     assets,
    ...     Timestamp('2014-01-03'),
    ...     Timestamp('2014-01-05'),
    ...     3,
    ...     0.5,
    ... )
    Float64Multiply(first_row=2, last_row=4, first_col=3, last_col=3, value=0.500000)
    """
    first_row, last_row, col = get_adjustment_locs(
        dates_index,
        assets_index,
        start_date,
        end_date,
        asset_id,
    )
    return cls(
        first_row=first_row,
        last_row=last_row,
        first_col=col,
        last_col=col,
        value=value,
    )

class Adjustment:
    """
    Base class for Adjustments.

    Subclasses should inherit and provide a `value` attribute and a `mutate`
    method.
    """

    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int):
        if not (0 <= first_row <= last_row):
            raise ValueError(
                'first_row must be in the range [0, last_row], got:'
                ' first_row=%s last_row=%s' % (
                    first_row,
                    last_row,
                ),
            )
        if not (0 <= first_col <= last_col):
            raise ValueError(
                'first_col must be in the range [0, last_col], got:'
                ' first_col=%s last_col=%s' % (
                    first_col,
                    last_col,
                ),
            )

        self.first_col = first_col
        self.last_col = last_col
        self.first_row = first_row
        self.last_row = last_row

    from_assets_and_dates = classmethod(_from_assets_and_dates)

    def __richcmp__(self, other, op: int):
        """
        Rich comparison method.  Only Equality is defined.
        """
        if type(self) != type(other):
            return NotImplemented

        return self._key() == other._key()

    def _key(self) -> tuple:
        """
        Comparison key
        """
        return (
            self.first_row,
            self.last_row,
            self.first_col,
            self.last_col,
            self.value,
        )

    def __reduce__(self):
        return type(self), self._key()


class AdjustmentKind(enum.Enum):
    MULTIPLY = 0
    ADD = 1
    OVERWRITE = 2


ADJUSTMENT_KIND_NAMES = {
    AdjustmentKind.MULTIPLY: 'MULTIPLY',
    AdjustmentKind.ADD: 'ADD',
    AdjustmentKind.OVERWRITE: 'OVERWRITE',
}


def _is_float(value):
    return isinstance(value, (float, float64))


def _is_datetime(value):
    return isinstance(value, (datetime64, Timestamp))


def _is_int(value):
    return isinstance(value, (int, int64))


def _is_obj(value):
    return isinstance(value, (bytes, unicode, type(None)))


def _is_bool(value):
    return isinstance(value, (bool, bool_, uint8))


def choose_adjustment_type(adjustment_kind: AdjustmentKind, value):
    """
    Make an adjustment object of the type appropriate for the given kind and
    value.

    Parameters
    ----------
    adjustment_kind : {ADD, MULTIPLY, OVERWRITE}
        The kind of adjustment to construct.
    value : object
        The value parameter to the adjustment.  Only floating-point values and
        datetime-like values are currently supported
    """
    if adjustment_kind in (AdjustmentKind.ADD, AdjustmentKind.MULTIPLY):
        if not _is_float(value):
            raise TypeError(
                "Can't construct %s Adjustment with value of type %r.\n"
                "ADD and MULTIPLY adjustments are only supported for "
                "floating point data." % (
                    ADJUSTMENT_KIND_NAMES[adjustment_kind],
                    type(value),
                )
            )
        return _float_adjustment_types[adjustment_kind]

    elif adjustment_kind == AdjustmentKind.OVERWRITE:
        if _is_float(value):
            return _float_adjustment_types[adjustment_kind]
        elif _is_datetime(value):
            return _datetime_adjustment_types[adjustment_kind]
        elif _is_bool(value):
            return _boolean_adjustment_types[adjustment_kind]
        elif _is_int(value):
            return _int_adjustment_types[adjustment_kind]
        elif _is_obj(value):
            return _object_adjustment_types[adjustment_kind]
        else:
            raise TypeError(
                "Don't know how to make overwrite "
                "adjustments for values of type %r." % type(value),
            )

    else:
        raise ValueError("Unknown adjustment type %d." % adjustment_kind)


def make_adjustment_from_indices(first_row: int,
                                 last_row: int,
                                 first_column: int,
                                 last_column: int,
                                 adjustment_kind: AdjustmentKind,
                                 value) -> Adjustment:
    """
    Make an Adjustment object from row/column indices into a baseline array.
    """
    type_ = choose_adjustment_type(adjustment_kind, value)
    # NOTE_SS: Cython appears to generate incorrect code here if values are
    # passed by name.  This is true even if cython.always_allow_keywords is
    # enabled.  Yay Cython.
    return type_(first_row, last_row, first_column, last_column, value)


def _choose_adjustment_type(adjustment_kind: AdjustmentKind,
                            value: Any) -> type:
    column_type = type(value)
    if adjustment_kind in (AdjustmentKind.ADD, AdjustmentKind.MULTIPLY):
        if column_type is np.float64:
            return _float_adjustment_types[adjustment_kind]
        else:
            raise TypeError(
                "Can't construct %s Adjustment with value of type %r.\n"
                "ADD and MULTIPLY adjustments are only supported for "
                "floating point data." % (
                    ADJUSTMENT_KIND_NAMES[adjustment_kind],
                    column_type,
                )
            )

    elif adjustment_kind == AdjustmentKind.OVERWRITE:
        if column_type is np.float64:
            return _float_adjustment_types[adjustment_kind]
        elif column_type is np.int64:
            return _int_adjustment_types[adjustment_kind]
        elif column_type is np.uint8:
            return _boolean_adjustment_types[adjustment_kind]
        elif column_type is object:
            return _object_adjustment_types[adjustment_kind]
        else:
            raise TypeError(
                "Don't know how to make overwrite "
                "adjustments for values of type %r." % type(value),
            )

    else:
        raise ValueError("Unknown adjustment type %d." % adjustment_kind)


def make_adjustment_from_indices_fused(first_row: int,
                                       last_row: int,
                                       first_column: int,
                                       last_column: int,
                                       adjustment_kind: AdjustmentKind,
                                       value: Any) -> Adjustment:
    """
    Make an Adjustment object from row/column indices into a baseline array.
    """
    type_ = _choose_adjustment_type(adjustment_kind, value)
    # NOTE_SS: Cython appears to generate incorrect code here if values are
    # passed by name.  This is true even if cython.always_allow_keywords is
    # enabled.  Yay Cython.
    return type_(first_row, last_row, first_column, last_column, value)


def make_adjustment_from_labels(dates_index: pd.DatetimeIndex,
                                assets_index: pd.RangeIndex,
                                start_date: pd.Timestamp,
                                end_date: pd.Timestamp,
                                asset_id: int,
                                adjustment_kind: AdjustmentKind,
                                value: Any):
    """
    Make an Adjustment object from date/asset labels into a labelled baseline
    array.
    """
    type_ = choose_adjustment_type(adjustment_kind, value)
    return type_.from_assets_and_dates(
        dates_index,
        assets_index,
        start_date,
        end_date,
        asset_id,
        value,
    )


def get_adjustment_locs(dates_index: pd.DatetimeIndex,
                        assets_index: pd.RangeIndex,
                        start_date: pd.Timestamp,
                        end_date: pd.Timestamp,
                        asset_id: int) -> tuple:
    """
    Compute indices suitable for passing to an Adjustment constructor.

    If the specified dates aren't in dates_index, we return the index of the
    first date **BEFORE** the supplied date.

    Example:

    >>> from pandas import date_range,  Timestamp
    >>> dates = date_range('2014-01-01', '2014-01-07')
    >>> assets = Int64Index(range(10))
    >>> get_adjustment_locs(
    ...     dates,
    ...     assets,
    ...     Timestamp('2014-01-03'),
    ...     Timestamp('2014-01-05'),
    ...     3,
    ... )
    (2, 4, 3)
    """

    # None or NaT signifies "All values before the end_date".
    if isnull(start_date):
        start_date_loc = 0
    else:
        # Location of earliest date on or after start_date.
        start_date_loc = dates_index.get_indexer([start_date], method='bfill')[0]

    return (
        start_date_loc,
        # Location of latest date on or before start_date.
        dates_index.get_indexer([end_date], method='ffill')[0],
        assets_index.get_indexer([asset_id])[0],  # Must be exact match.
    )




class Float64Adjustment(Adjustment):
    """
    Base class for adjustments that operate on Float64 data.
    """

    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 value: float):
        super(Float64Adjustment, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
        )
        self.value = value

    from_assets_and_dates = classmethod(_from_assets_and_dates)

    def __repr__(self):
        return (
                "%s(first_row=%d, last_row=%d,"
                " first_col=%d, last_col=%d, value=%f)" % (
                    type(self).__name__,
                    self.first_row,
                    self.last_row,
                    self.first_col,
                    self.last_col,
                    self.value,
                )
        )


class Float64Multiply(Float64Adjustment):
    """
    An adjustment that multiplies by a float.

    Example
    -------

    >>> import numpy as np
    >>> arr = np.arange(9, dtype=float).reshape(3, 3)
    >>> arr
    array([[ 0.,  1.,  2.],
           [ 3.,  4.,  5.],
           [ 6.,  7.,  8.]])

    >>> adj = Float64Multiply(
    ...     first_row=1,
    ...     last_row=2,
    ...     first_col=1,
    ...     last_col=2,
    ...     value=4.0,
    ... )
    >>> adj.mutate(arr)
    >>> arr
    array([[  0.,   1.,   2.],
           [  3.,  16.,  20.],
           [  6.,  28.,  32.]])
    """

    def mutate(self, data):
        value = self.value

        # last_col + 1 because last_col should also be affected.
        for col in range(self.first_col, self.last_col + 1):
            # last_row + 1 because last_row should also be affected.
            for row in range(self.first_row, self.last_row + 1):
                data[row, col] *= value


class Float64Overwrite(Float64Adjustment):
    """
    An adjustment that overwrites with a float.

    Example
    -------

    >>> import numpy as np
    >>> arr = np.arange(9, dtype=float).reshape(3, 3)
    >>> arr
    array([[ 0.,  1.,  2.],
           [ 3.,  4.,  5.],
           [ 6.,  7.,  8.]])

    >>> adj = Float64Overwrite(
    ...     first_row=1,
    ...     last_row=2,
    ...     first_col=1,
    ...     last_col=2,
    ...     value=0.0,
    ... )
    >>> adj.mutate(arr)
    >>> arr
    array([[ 0.,  1.,  2.],
           [ 3.,  0.,  0.],
           [ 6.,  0.,  0.]])
    """

    def mutate(self, data):
        value = self.value

        # last_col + 1 because last_col should also be affected.
        for col in range(self.first_col, self.last_col + 1):
            # last_row + 1 because last_row should also be affected.
            for row in range(self.first_row, self.last_row + 1):
                data[row, col] = value


class ArrayAdjustment(Adjustment):
    """
    Base class for ArrayAdjustments.

    Subclasses should inherit and provide a `values` attribute and a `mutate`
    method.
    """

    def __repr__(self):
        return (
                "%s(first_row=%d, last_row=%d,"
                " first_col=%d, last_col=%d, values=%s)" % (
                    type(self).__name__,
                    self.first_row,
                    self.last_row,
                    self.first_col,
                    self.last_col,
                    asarray(self.values),
                )
        )


class Float641DArrayOverwrite(ArrayAdjustment):
    """
    An adjustment that overwrites subarrays with a value for each subarray.

    Example
    -------

    >>> import numpy as np
    >>> arr = np.arange(25, dtype=float).reshape(5, 5)
    >>> arr
    array([[  0.,   1.,   2.,   3.,   4.],
           [  5.,   6.,   7.,   8.,   9.],
           [ 10.,  11.,  12.,  13.,  14.],
           [ 15.,  16.,  17.,  18.,  19.],
           [ 20.,  21.,  22.,  23.,  24.]])
    >>> adj = Float641DArrayOverwrite(
    ...     row_start=0,
    ...     row_end=3,
    ...     column_start=0,
    ...     column_end=0,
    ...     values=np.array([1, 2, 3, 4]),
    )
    >>> adj.mutate(arr)
    >>> arr
    array([[  1.,   1.,   2.,   3.,   4.],
           [  2.,   6.,   7.,   8.,   9.],
           [ 3.,  11.,  12.,  13.,  14.],
           [ 4.,  16.,  17.,  18.,  19.],
           [ 20.,  21.,  22.,  23.,  24.]])
    """

    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 values: np.ndarray):
        super(Float641DArrayOverwrite, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
        )
        if last_row + 1 - first_row != len(values):
            raise ValueError(
                "Mismatch: got %d values for rows starting at index %d and "
                "ending at index %d." % (len(values), first_row, last_row)
            )
        self.values = values

    def mutate(self, data):
        values = self.values
        for col in range(self.first_col, self.last_col + 1):
            for i, row in enumerate(range(self.first_row, self.last_row + 1)):
                data[row, col] = values[i]


class Datetime641DArrayOverwrite(ArrayAdjustment):
    """
    An adjustment that overwrites subarrays with a value for each subarray.

    Example
    -------

    >>> import numpy as np; import pandas as pd
    >>> dts = pd.date_range('2014', freq='D', periods=9, tz='UTC')
    >>> arr = dts.values.reshape(3, 3)
    >>> arr == np.datetime64(0, 'ns')
    array([[False, False, False],
       [False, False, False],
       [False, False, False]], dtype=bool)
    >>> adj = Datetime641DArrayOverwrite(
    ...           first_row=1,
    ...           last_row=2,
    ...           first_col=1,
    ...           last_col=2,
    ...           values=np.array([
    ...               np.datetime64(0, 'ns'),
    ...               np.datetime64(1, 'ns')
    ...           ])
    ...       )
    >>> adj.mutate(arr.view(np.int64))
    >>> arr == np.datetime64(0, 'ns')
    array([[False, False, False],
       [False,  True,  True],
       [False, False, False]], dtype=bool)
    >>> arr == np.datetime64(1, 'ns')
    array([[False, False, False],
       [False, False, False],
       [False,  True,  True]], dtype=bool)
    """

    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 values):
        super(Datetime641DArrayOverwrite, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
        )
        if last_row + 1 - first_row != len(values):
            raise ValueError("Mismatch: got %d values for rows starting at"
                             " index %d and ending at index %d." % (
                                 len(values), first_row, last_row)
                             )
        self.values = asarray([datetime_to_int(value) for value in values])

    def mutate(self, data):
        values = self.values
        for col in range(self.first_col, self.last_col + 1):
            for i, row in enumerate(range(self.first_row, self.last_row + 1)):
                data[row, col] = values[i]

    def __reduce__(self):
        return type(self), (
            self.first_row,
            self.last_row,
            self.first_col,
            self.last_col,
            self.values.view('datetime64[ns]'),
        )


class Object1DArrayOverwrite(ArrayAdjustment):
    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 values):
        super(Object1DArrayOverwrite, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
        )
        if last_row + 1 - first_row != len(values):
            raise ValueError(
                "Mismatch: got %d values for rows starting at index %d and "
                "ending at index %d." % (len(values), first_row, last_row)
            )
        self.values = values[:, None]

    def mutate(self, data):
        # data is an object here because this is intended to be used with a
        # `ziplime.lib.LabelArray`.

        data[
        self.first_row:self.last_row + 1,
        self.first_col:self.last_col + 1,
        ] = self.values


class Boolean1DArrayOverwrite(ArrayAdjustment):
    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 values):
        if values.dtype.kind != 'b':
            raise TypeError('dtype is not bool, got: %r' % values.dtype)

        super(Boolean1DArrayOverwrite, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
        )

        if last_row + 1 - first_row != len(values):
            raise ValueError(
                "Mismatch: got %d values for rows starting at index %d and "
                "ending at index %d." % (len(values), first_row, last_row)
            )
        self.values = values.view('uint8')

    def mutate(self, data):
        values = self.values
        for col in range(self.first_col, self.last_col + 1):
            for i, row in enumerate(range(self.first_row, self.last_row + 1)):
                data[row, col] = values[i]

    def __repr__(self):
        return (
                "%s(first_row=%d, last_row=%d,"
                " first_col=%d, last_col=%d, value=%r)" % (
                    type(self).__name__,
                    self.first_row,
                    self.last_row,
                    self.first_col,
                    self.last_col,
                    self.value.view('?'),
                )
        )

    def __reduce__(self):
        return type(self), (
            self.first_row,
            self.last_row,
            self.first_col,
            self.last_col,
            self.value.view('?'),
        )


class Float64Add(Float64Adjustment):
    """
    An adjustment that adds a float.

    Example
    -------

    >>> import numpy as np
    >>> arr = np.arange(9, dtype=float).reshape(3, 3)
    >>> arr
    array([[ 0.,  1.,  2.],
           [ 3.,  4.,  5.],
           [ 6.,  7.,  8.]])

    >>> adj = Float64Add(
    ...     first_row=1,
    ...     last_row=2,
    ...     first_col=1,
    ...     last_col=2,
    ...     value=1.0,
    ... )
    >>> adj.mutate(arr)
    >>> arr
    array([[ 0.,  1.,  2.],
           [ 3.,  5.,  6.],
           [ 6.,  8.,  9.]])
    """

    def mutate(self, data):
        value = self.value

        # last_col + 1 because last_col should also be affected.
        for col in range(self.first_col, self.last_col + 1):
            # last_row + 1 because last_row should also be affected.
            for row in range(self.first_row, self.last_row + 1):
                data[row, col] += value


class _Int64Adjustment(Adjustment):
    """
    Base class for adjustments that operate on integral data.

    This is private because we never actually operate on integers as data, but
    we use integer arrays to represent datetime and timedelta data.
    """

    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 value: int):
        super(_Int64Adjustment, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
        )
        self.value = value

    def __repr__(self):
        return (
                "%s(first_row=%d, last_row=%d,"
                " first_col=%d, last_col=%d, value=%d)" % (
                    type(self).__name__,
                    self.first_row,
                    self.last_row,
                    self.first_col,
                    self.last_col,
                    self.value,
                )
        )


class Int64Overwrite(_Int64Adjustment):
    """
    An adjustment that overwrites with an int.

    Example
    -------

    >>> import numpy as np
    >>> arr = np.arange(9, dtype=int).reshape(3, 3)
    >>> arr
    array([[ 0,  1,  2],
           [ 3,  4,  5],
           [ 6,  7,  8]])

    >>> adj = Int64Overwrite(
    ...     first_row=1,
    ...     last_row=2,
    ...     first_col=1,
    ...     last_col=2,
    ...     value=0,
    ... )
    >>> adj.mutate(arr)
    >>> arr
    array([[ 0,  1,  2],
           [ 3,  0,  0],
           [ 6,  0,  0]])
    """

    def mutate(self, data):
        value = self.value

        # last_col + 1 because last_col should also be affected.
        for col in range(self.first_col, self.last_col + 1):
            # last_row + 1 because last_row should also be affected.
            for row in range(self.first_row, self.last_row + 1):
                data[row, col] = value


def datetime_to_int(datetimelike):
    """
    Coerce a datetime-like object to the int format used by AdjustedArrays of
    Datetime64 type.
    """
    if isinstance(datetimelike, Timestamp):
        return datetimelike.value

    if not isinstance(datetimelike, datetime64):
        raise TypeError("Expected datetime64, got %s" % type(datetimelike))

    elif datetimelike.dtype.name != 'datetime64[ns]':
        raise TypeError(
            "Expected datetime64[ns], got %s",
            datetimelike.dtype.name,
        )

    return datetimelike.astype(int64)


class Datetime64Adjustment(_Int64Adjustment):
    """
    Base class for adjustments that operate on Datetime64 data.

    Notes
    -----
    Numpy stores datetime64 values in arrays of type int64.  There's no
    straightforward way to work with statically-typed datetime64 data, so
    instead we work with int64 values everywhere, and we do validation/coercion
    at API boundaries.
    """

    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 value):
        super(Datetime64Adjustment, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
            value=datetime_to_int(value),
        )

    def __repr__(self):
        return (
                "%s(first_row=%d, last_row=%d,"
                " first_col=%d, last_col=%d, value=%r)" % (
                    type(self).__name__,
                    self.first_row,
                    self.last_row,
                    self.first_col,
                    self.last_col,
                    datetime64(self.value, 'ns'),
                )
        )

    def __reduce__(self):
        return type(self), (
            self.first_row,
            self.last_row,
            self.first_col,
            self.last_col,
            datetime64(self.value, 'ns'),
        )


class Datetime64Overwrite(Datetime64Adjustment):
    """
    An adjustment that overwrites with a datetime.

    This operates on int64 data which should be interpreted as nanoseconds
    since the epoch.

    Example
    -------

    >>> import numpy as np; import pandas as pd
    >>> dts = pd.date_range('2014', freq='D', periods=9, tz='UTC')
    >>> arr = dts.values.reshape(3, 3)
    >>> arr == np.datetime64(0, 'ns')
    array([[False, False, False],
           [False, False, False],
           [False, False, False]], dtype=bool)
    >>> adj = Datetime64Overwrite(
    ...     first_row=1,
    ...     last_row=2,
    ...     first_col=1,
    ...     last_col=2,
    ...     value=np.datetime64(0, 'ns'),
    ... )
    >>> adj.mutate(arr.view(np.int64))
    >>> arr == np.datetime64(0, 'ns')
    array([[False, False, False],
           [False,  True,  True],
           [False,  True,  True]], dtype=bool)
    """

    def mutate(self, data):
        value = self.value

        # last_col + 1 because last_col should also be affected.
        for col in range(self.first_col, self.last_col + 1):
            # last_row + 1 because last_row should also be affected.
            for row in range(self.first_row, self.last_row + 1):
                data[row, col] = value


class _ObjectAdjustment(Adjustment):
    """
    Base class for adjustments that operate on arbitrary objects.

    We use only this for categorical data, where our data buffer is an array of
    indices into an array of unique Python string objects.
    """

    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 value):
        super(_ObjectAdjustment, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
        )
        self.value = value

    def __repr__(self):
        return (
                "%s(first_row=%d, last_row=%d,"
                " first_col=%d, last_col=%d, value=%r)" % (
                    type(self).__name__,
                    self.first_row,
                    self.last_row,
                    self.first_col,
                    self.last_col,
                    self.value,
                )
        )


class ObjectOverwrite(_ObjectAdjustment):

    def mutate(self, data):
        # data is an object here because this is intended to be used with a
        # `ziplime.lib.LabelArray`.

        # We don't do this in a loop because we only want to look up the label
        # code in the array's categories once.
        data.set_scalar(
            (
                slice(self.first_row, self.last_row + 1),
                slice(self.first_col, self.last_col + 1),
            ),
            self.value,
        )


class BooleanAdjustment(Adjustment):
    """
    Base class for adjustments that operate on boolean data.

    Notes
    -----
    Numpy stores boolean values in arrays of type uint8.  There's no
    straightforward way to work with statically-typed boolean data, so
    instead we work with uint8 values everywhere, and we do validation/coercion
    at API boundaries.
    """

    def __init__(self,
                 first_row: int,
                 last_row: int,
                 first_col: int,
                 last_col: int,
                 value: int):  # bint -> boolean int

        super(BooleanAdjustment, self).__init__(
            first_row=first_row,
            last_row=last_row,
            first_col=first_col,
            last_col=last_col,
        )
        self.value = value

    def __repr__(self):
        return (
                "%s(first_row=%d, last_row=%d,"
                " first_col=%d, last_col=%d, value=%r)" % (
                    type(self).__name__,
                    self.first_row,
                    self.last_row,
                    self.first_col,
                    self.last_col,
                    bool(self.value),
                )
        )

    def __reduce__(self):
        return type(self), (
            self.first_row,
            self.last_row,
            self.first_col,
            self.last_col,
            bool(self.value),
        )


class BooleanOverwrite(BooleanAdjustment):
    """
    An adjustment that overwrites with a boolean.

    This operates on uint8 data.
    """

    def mutate(self, data):
        value = self.value

        # last_col + 1 because last_col should also be affected.
        for col in range(self.first_col, self.last_col + 1):
            # last_row + 1 because last_row should also be affected.
            for row in range(self.first_row, self.last_row + 1):
                data[row, col] = value


_float_adjustment_types = {
    AdjustmentKind.ADD: Float64Add,
    AdjustmentKind.MULTIPLY: Float64Multiply,
    AdjustmentKind.OVERWRITE: Float64Overwrite,
}
_datetime_adjustment_types = {
    AdjustmentKind.OVERWRITE: Datetime64Overwrite,
}
_object_adjustment_types = {
    AdjustmentKind.OVERWRITE: ObjectOverwrite,
}
_int_adjustment_types = {
    AdjustmentKind.OVERWRITE: Int64Overwrite,
}
_boolean_adjustment_types = {
    AdjustmentKind.OVERWRITE: BooleanOverwrite,
}



import datetime

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.slippage.slippage_model import SlippageModel


class NoSlippage(SlippageModel):
    """A slippage model where all orders fill immediately and completely at the
    current close price.

    Notes
    -----
    This is primarily used for testing.
    """

    @staticmethod
    def process_order(exchange: Exchange, dt:datetime.datetime, order):
        return (
            data.current(order.asset, "close"),
            order.amount,
        )




from ziplime.finance.controls.trading_control import TradingControl


class MaxOrderCount(TradingControl):
    """TradingControl representing a limit on the number of orders that can be
    placed in a given trading day.
    """

    def __init__(self, on_error, max_count):

        super(MaxOrderCount, self).__init__(on_error, max_count=max_count)
        self.orders_placed = 0
        self.max_count = max_count
        self.current_date = None

    def validate(self, asset, amount, portfolio, algo_datetime, algo_current_data):
        """Fail if we've already placed self.max_count orders today."""
        algo_date = algo_datetime.date()

        # Reset order count if it's a new day.
        if self.current_date and self.current_date != algo_date:
            self.orders_placed = 0
        self.current_date = algo_date

        if self.orders_placed >= self.max_count:
            self.handle_violation(asset, amount, algo_datetime)
        self.orders_placed += 1



import datetime

import aiofiles.os
from pathlib import Path
from typing import Any, Sequence, Literal, Self

import structlog
from polars import CredentialProviderFunction, Expr
from polars._typing import ParquetCompression
import polars as pl

from ziplime.constants.data_type import DataType
from ziplime.constants.period import Period
from ziplime.data.domain.data_bundle import DataBundle
from ziplime.data.services.bundle_storage import BundleStorage


class FileSystemParquetBundleStorage(BundleStorage):

    def __init__(self,
                 base_data_path: str,
                 compression: ParquetCompression = "brotli",
                 compression_level: int | None = None,
                 statistics: bool | str | dict[str, bool] = True,
                 row_group_size: int | None = None,
                 data_page_size: int | None = None,
                 use_pyarrow: bool = False,
                 pyarrow_options: dict[str, Any] | None = None,
                 partition_by: str | Sequence[str] | None = None,
                 partition_chunk_size_bytes: int = 4_294_967_296,
                 storage_options: dict[str, Any] | None = None,
                 credential_provider: (
                         CredentialProviderFunction | Literal["auto"] | None
                 ) = "auto", ):
        super().__init__()
        self.base_data_path = base_data_path
        self.compression = compression
        self.compression_level = compression_level
        self.statistics = statistics
        self.row_group_size = row_group_size
        self.data_page_size = data_page_size
        self.use_pyarrow = use_pyarrow
        self.pyarrow_options = pyarrow_options
        self.partition_by = partition_by
        self.partition_chunk_size_bytes = partition_chunk_size_bytes
        self.storage_options = storage_options
        self._logger = structlog.get_logger(__name__)

    async def store_bundle(self, data_bundle: DataBundle):
        # we need here to know ehere to store bundle, and info is in bundle metadata
        bundle_path = self.get_data_bundle_path(data_bundle=data_bundle)
        await aiofiles.os.makedirs(bundle_path.parent, exist_ok=True)
        data_bundle.data.write_parquet(bundle_path, compression=self.compression,
                                       compression_level=10,
                                       statistics=self.statistics, row_group_size=self.row_group_size,
                                       data_page_size=self.data_page_size, use_pyarrow=self.use_pyarrow,
                                       pyarrow_options=self.pyarrow_options, partition_by=self.partition_by,
                                       partition_chunk_size_bytes=self.partition_chunk_size_bytes,
                                       storage_options=self.storage_options)

    async def load_data_bundle(self, data_bundle: DataBundle,
                               symbols: list[str] | None = None,
                               start_date: datetime.datetime | None = None,
                               end_date: datetime.datetime | None = None,
                               frequency: datetime.timedelta | Period | None = None,
                               start_auction_delta: datetime.timedelta = None,
                               end_auction_delta: datetime.timedelta = None,
                               aggregations: list[pl.Expr] = None
                               ) -> pl.DataFrame:
        self._logger.info(
            f"Loading data bundle {data_bundle.name} start_date={start_date}, end_date={end_date},"
            f" version={data_bundle.version}, frequency={frequency}")
        bundle_path = self.get_data_bundle_path(data_bundle=data_bundle)
        filters = []
        pl_parquet = pl.scan_parquet(bundle_path)
        if symbols is not None:
            filters.append(pl.col("symbol").is_in(symbols))
        if start_date is not None:
            filters.append(pl.col("date") >= start_date)
        if end_date is not None:
            filters.append(pl.col("date") <= end_date)

        if filters:
            pl_parquet = pl_parquet.filter(*filters)

        if frequency is not None:

            if data_bundle.aggregation_specification:
                pl_parquet = pl_parquet.group_by_dynamic(
                    index_column="date", every=frequency, by="sid").agg(
                    pl.col(field).last() for field in pl_parquet.collect_schema().names() if
                    field not in ('sid', 'date')
                )
            else:
                if start_auction_delta is not None:
                    pl_parquet = pl_parquet.filter(
                        pl.col("date") >= (
                                pl.col("date").min().over(pl.col("date").dt.date()) + pl.duration(
                            seconds=start_auction_delta.total_seconds())
                        )
                    )
                if end_auction_delta is not None:
                    pl_parquet = pl_parquet.filter(pl.col("date") <= (
                            pl.col("date").max().over(pl.col("date").dt.date()) - pl.duration(
                        seconds=end_auction_delta.total_seconds())
                    ))

                if not aggregations and data_bundle.data_type == DataType.MARKET_DATA:
                    aggregations = [
                        pl.col("open").first(),
                        pl.col("high").max(),
                        pl.col("low").min(),
                        pl.col("volume").sum()
                    ]
                if aggregations:
                    aggregation_columns = {aggregation.meta.output_name() for aggregation in aggregations}
                    # default aggregation is last()
                    missing_aggregation_columns = [
                        pl.col(col).last() for col in pl_parquet.collect_schema().names()
                        if col not in aggregation_columns and col not in ('sid', 'date')
                    ]
                    pl_parquet = pl_parquet.group_by_dynamic(
                        index_column="date", every=frequency, by="sid").agg(*aggregations, *missing_aggregation_columns)
                else:
                    pl_parquet = pl_parquet.group_by_dynamic(
                        index_column="date", every=frequency, by="sid").agg(
                        pl.col(field).last() for field in pl_parquet.collect_schema().names() if
                        field not in ('sid', 'date')
                    )
        return pl_parquet.sort(["sid", "date"]).collect()

    @classmethod
    async def from_json(cls, data: dict[str, Any]) -> Self:
        return cls(base_data_path=data["base_data_path"])

    def get_data_bundle_path(self, data_bundle: DataBundle) -> Path:
        return Path(self.base_data_path, "data_bundle", data_bundle.name, data_bundle.version, f"data.parquet")

    async def to_json(self, data_bundle: DataBundle) -> dict[str, Any]:
        return {
            "base_data_path": self.base_data_path,
        }



import datetime
import operator as op
from typing import Any

import pandas as pd
from exchange_calendars import ExchangeCalendar

from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger

from ziplime.sources.benchmark_source import BenchmarkSource


class StartOfPeriodLedgerField:
    """Keep track of the value of a ledger field at the start of the period.

    Parameters
    ----------
    ledger_field : str
        The ledger field to read.
    packet_field : str, optional
        The name of the field to populate in the packet. If not provided,
        ``ledger_field`` will be used.
    """

    def __init__(self, ledger_field, packet_field=None):
        self._get_ledger_field = op.attrgetter(ledger_field)
        if packet_field is None:
            self._packet_field = ledger_field.rsplit(".", 1)[-1]
        else:
            self._packet_field = packet_field
        self._previous_day = 0.0

    def start_of_simulation(
            self, ledger: Ledger, emission_rate: datetime.timedelta, trading_calendar: ExchangeCalendar,
            sessions: pd.DatetimeIndex, benchmark_source: BenchmarkSource
    ):
        self._start_of_simulation = self._get_ledger_field(ledger)

    def start_of_session(self, ledger, session, exchanges: dict[str, Exchange]):
        self._previous_day = self._get_ledger_field(ledger)

    def _end_of_period(self, sub_field, packet, ledger):
        # if not hasattr(self, "_previous_day"):
        #     self._previous_day = self._get_ledger_field(ledger)
        packet_field = self._packet_field
        packet["cumulative_perf"][packet_field] = self._start_of_simulation
        packet[sub_field][packet_field] = self._previous_day

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        self._end_of_period("minute_perf", packet, ledger)

    def end_of_session(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                       exchanges: dict[str, Exchange]):
        self._end_of_period("daily_perf", packet, ledger)



#
# Copyright 2014 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note that part of the API is implemented in TradingAlgorithm as
# methods (e.g. order). These are added to this namespace via the
# decorator ``api_method`` inside of algorithm.py.
from .finance.asset_restrictions import (
    Restriction,
    StaticRestrictions,
    HistoricalRestrictions,
    RESTRICTION_STATES,
)
from .finance import commission, execution, slippage, cancel_policy
from .finance.cancel_policy import NeverCancel, EODCancel
from .finance.slippage.fixed_basis_points_slippage import FixedBasisPointsSlippage
from .finance.slippage.fixed_slippage import FixedSlippage
from .finance.slippage.volume_share_slippage import VolumeShareSlippage

from .utils import math_utils, events
from .utils.events import calendars, date_rules, time_rules

__all__ = [
    "EODCancel",
    "FixedSlippage",
    "FixedBasisPointsSlippage",
    "NeverCancel",
    "VolumeShareSlippage",
    "Restriction",
    "StaticRestrictions",
    "HistoricalRestrictions",
    "RESTRICTION_STATES",
    "cancel_policy",
    "commission",
    "date_rules",
    "events",
    "execution",
    "math_utils",
    "slippage",
    "time_rules",
    "calendars",
]



import polars as pl


class AssetDataSource:

    def __init__(self):
        pass

    async def get_assets(self, **kwargs) -> pl.DataFrame:
        pass

    async def get_constituents(self, index: str) -> pl.DataFrame: ...

    async def search_assets(self, query: str, **kwargs) -> pl.DataFrame: ...



#
# Copyright 2013 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from decimal import Decimal
import math

from numpy import isnan


def tolerant_equals(a, b, atol=10e-7, rtol=10e-7, equal_nan=False):
    """Check if a and b are equal with some tolerance.

    Parameters
    ----------
    a, b : float
        The floats to check for equality.
    atol : float, optional
        The absolute tolerance.
    rtol : float, optional
        The relative tolerance.
    equal_nan : bool, optional
        Should NaN compare equal?

    See Also
    --------
    numpy.isclose

    Notes
    -----
    This function is just a scalar version of numpy.isclose for performance.
    See the docstring of ``isclose`` for more information about ``atol`` and
    ``rtol``.
    """
    if equal_nan and isnan(a) and isnan(b):
        return True
    return math.isclose(a, b, rel_tol=rtol, abs_tol=atol)
    # return math.fabs(a - b) <= (atol + rtol * math.fabs(b))


try:
    # fast versions
    import bottleneck as bn

    nanmean = bn.nanmean
    nanstd = bn.nanstd
    nansum = bn.nansum
    nanmax = bn.nanmax
    nanmin = bn.nanmin
    nanargmax = bn.nanargmax
    nanargmin = bn.nanargmin
    nanmedian = bn.nanmedian
except ImportError:
    # slower numpy
    import numpy as np

    nanmean = np.nanmean
    nanstd = np.nanstd
    nansum = np.nansum
    nanmax = np.nanmax
    nanmin = np.nanmin
    nanargmax = np.nanargmax
    nanargmin = np.nanargmin
    nanmedian = np.nanmedian


def round_if_near_integer(a, epsilon=1e-4):
    """
    Round a to the nearest integer if that integer is within an epsilon
    of a.
    """
    if abs(a - round(a)) <= epsilon:
        return round(a)
    else:
        return a


def number_of_decimal_places(n):
    """
    Compute the number of decimal places in a number.

    Examples
    --------
    >>> number_of_decimal_places(1)
    0
    >>> number_of_decimal_places(3.14)
    2
    >>> number_of_decimal_places('3.14')
    2
    """
    decimal = Decimal(str(n))
    return -decimal.as_tuple().exponent



from dataclasses import dataclass

from ziplime.assets.entities.asset import Asset


@dataclass(frozen=True)
class Commodity(Asset):
    # mapping exchange code -> symbol
    symbol_mapping: dict[str, str]

    def get_symbol_by_exchange(self, exchange_name: str) -> str | None:
        return self.symbol_mapping.get(exchange_name, None)
